# -*- coding: utf-8 -*-
#  privacyIDEA is a fork of LinOTP
#  May 08, 2014 Cornelius Kölbel
#  License:  AGPLv3
#  contact:  http://www.privacyidea.org
#
#  Copyright (C) 2010 - 2014 LSE Leading Security Experts GmbH
#  License:  AGPLv3
#  contact:  http://www.linotp.org
#            http://www.lsexperts.de
#            linotp@lsexperts.de
#
# This code is free software; you can redistribute it and/or
# modify it under the terms of the GNU AFFERO GENERAL PUBLIC LICENSE
# License as published by the Free Software Foundation; either
# version 3 of the License, or any later version.
#
# This code is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU AFFERO GENERAL PUBLIC LICENSE for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
'''     
  Description:  functional tests
                
  Dependencies: -

'''
import sys
import ConfigParser
import logging
from privacyidea.tests import TestController, url
import copy

import json

import traceback
import hashlib
import base64

from socket import gethostname


log = logging.getLogger(__name__)

import ldap



class LDAP(object):

    def __init__(self, ldapurl=None):

        self.lobj = None
        self.binddn = None
        self.bindpw = None
        self.ldapserverip = None
        self.ldapuri = None
        self.base = None

        if ldapurl is not None:
            self.ldapurl = ldapurl

            (part, _split, hostname) = self.ldapurl.partition('@')
            if part.startswith('ldap://'):
                offstring = 'ldap://'
                self.ldapuri = 'ldap://'
            if part.startswith('ldaps://'):
                offstring = 'ldaps://'
                self.ldapuri = 'ldaps://'

            userpass = part[len(offstring):]
            posCol = userpass.index(':')

            self.binddn = userpass[:posCol]
            self.bindpw = userpass[posCol + 1:]
            self.ldapserverip = hostname
            self.ldapuri = self.ldapuri + hostname

            baseIndex = self.binddn.index('dc')
            self.base = self.binddn[baseIndex:]

        return


    def loadConfig(self, config, conf="", bindpw=None, binddn=None):
        '''
            loadConfig - load the config for the resolver
        '''

        self.conf = conf
        if conf is not None:
            conf = '.' + conf

        if binddn is not None:
            self.binddn = binddn
        else:
            self.binddn = config.get("ldapresolver.BINDDN" + conf)

        if bindpw is not None:
            self.bindpw = bindpw
        else:
            self.bindpw = config.get("ldapresolver.BINDPW" + conf)


        self.filter = config.get("ldapresolver.LDAPFILTER" + conf, '')
        self.searchfilter = config.get("ldapresolver.LDAPSEARCHFILTER" + conf)
        self.ldapuri = config.get("ldapresolver.LDAPURI" + conf)
        self.base = config.get("ldapresolver.LDAPBASE" + conf)

        self.loginnameattribute = config.get("ldapresolver.LOGINNAMEATTRIBUTE" + conf)
        #userinfo           = config.get("ldapresolver.USERINFO"+conf)
        #timeout            = config.get("ldapresolver.TIMEOUT"+conf)

        self.uidType = config.get("ldapresolver.UIDTYPE" + conf)
        if self.uidType is None or self.uidType.strip() == "":
            self.uidType = "DN"

        sizelimit = config.get("ldapresolver.SIZELIMIT" + conf)
        try:
            self.sizelimit = int(sizelimit)
        except ValueError:
            self.sizelimit = 50
        except TypeError:
            self.sizelimit = 50


        return self


    def bind(self, binddn=None, bindpw=None):
        """ bind()  - this function starts an ldap conncetion
        """
        uri = ''
        #i   = 0


        if binddn is None:
            binddn = self.binddn

        if bindpw is None:
            bindpw = self.bindpw


        urilist = self.ldapuri.split(',')

        if self.lobj is None:
            for uri in urilist:
                try:
                    log.debug("[bind] LDAP: Try to bind to %s", uri)
                    l_obj = ldap.initialize(uri, trace_level=0)
                    l_obj.simple_bind_s(binddn, bindpw)
                    self.lobj = l_obj
                    break
                except ldap.LDAPError as e:
                    log.error("[bind] LDAP error: %r" % e)
                    log.error("[bind] LDAPURI   : %r" % uri)
                    log.error("[bind] %s" % traceback.format_exc())
                    raise Exception(e)

        return self.lobj


    def unbind(self):
        """ unbind() - this function frees the ldap connection
        """
        try:
            if self.lobj is not None:
                self.lobj.unbind_s()

        except ldap.LDAPError as e:
            log.error("[unbind] LDAP error: %r" % e)

        self.lobj = None
        return ""

    def addModlist(self, obj):
        '''
            transform dict to tuple
            -ldap requires the data as tuples
        '''
        record = []
        for k in obj:
            data = obj.get(k)
            record.append((k, data))

        return record

    def addUser(self, dn=None,
                objectclass=None, cn=None, uid=None,
                mail=None, givenName=None, sn=None,
                postalAddress=None, postalCode=None, l=None,
                homePhone=None, mobile=None,
                userPassword=None):
        '''
            # The dn of our new entry/object
            dn="cn=Sigmund Freud,ou=Lab,ou=people,dc=crypton,dc=info"

            """
            dn: cn=Werner Braun,ou=Lab,dc=crypton,dc=info
            # which schema to use
            objectClass: top
            objectClass: person
            objectClass: organizationalPerson
            objectClass: inetOrgPerson
            """

            """
    '''


        attrs = {}
        if dn is None:
            raise Exception('missing requred dn')

        if objectclass is None:
            attrs['objectclass'] = ['top', 'person', 'organizationalPerson', 'inetOrgPerson', 'uidObject']
        else:
            attrs['objectclass'] = objectclass.split(',')

        if cn is not None:
            attrs['cn'] = cn
        if givenName is not None:
            attrs['givenName'] = givenName
        if uid is not None:
            attrs['uid'] = uid
        elif cn is not None:
            attrs['uid'] = cn

        if sn is not None:
            attrs['sn'] = sn
        if postalAddress is not None:
            attrs['postalAddress'] = postalAddress
        if mail is not None:
            attrs['mail'] = mail
        if postalCode is not None:
            attrs['postalCode'] = postalCode
        if l is not None:
            attrs['l'] = l
        if homePhone is not None:
            attrs['homePhone'] = homePhone
        if mobile is not None:
            attrs['mobile'] = mobile

        if userPassword is not None:
            attrs['userPassword'] = userPassword


        self.lobj = self.bind()
        if self.lobj is None:
            raise Exception('LDAP Bind failed!')

        ldif = self.addModlist(attrs)
        self.lobj.add_s(dn, ldif)


        log.debug('userAdded done!')

        return

    def delete(self, dn):

        self.lobj = self.bind()
        if self.lobj is None:
            raise Exception('LDAP Bind failed!')

        self.lobj.delete_s(dn)

        log.debug('delete done!')

        return


    def addOu(self, dn, ou):
        attrs = {}
        attrs['objectclass'] = ['top', 'organizationalUnit']
        attrs['ou'] = ou

        self.lobj = self.bind()
        if self.lobj is None:
            raise Exception('LDAP Bind failed!')

        ldif = self.addModlist(attrs)
        self.lobj.add_s(dn, ldif)

        return

class TestLDAP(TestController):

    users = [
        {'dn'       : 'cn=Werner Braun,',
        'cn'        : 'Werner Braun',
        'givenName' : 'Werner',
        'uid'       : 'wärner',
        'sn'        : 'Braun',
        'mail'      : 'w.Braun@lsexperts.de',
        'postalAddress': 'My Way 5',
        'postalCode': '64297',
        'l'         : 'Darmstadt',
        'homePhone' : '06151 /123 456 789',
        'mobile'    : '0179 / 123 123 123', },

        {'dn'       : 'cn=Jörg Preuße,',
        'cn'        : 'Jörg Preuße',
        'givenName' : 'Preußer',
        'uid'       : 'jörgi',
        'sn'        : 'Jörg Preuße',
        'mail'      : 'Joerg.Preusse@lsexperts.de',
        'postalAddress': 'Lindenstraße 13',
        'postalCode': '24297',
        'l'         : 'Am Hintersee',
        'homePhone' : '06151 /123 456 789',
        'mobile'    : '0179 / 123 123 123', },

    ]





    def setupLDAPResolver(self, name='ldapUsers', ip='serverIp', base='ou=users,dc=crypton,dc=info', uidType='entryUUID', bindpw='test123!', binddn='cn=admin,dc=crypton,dc=info'):
        '''
            'LDAPBASE'              : 'ou=users,dc=crypton,dc=info',
            'LDAPBASE'              : 'ou=users,dc=crypton,dc=info',
        '''


        ldapDef = {
            'name'                  : name,
            'type'                  : 'ldapresolver',
            'LDAPURI'               : 'ldap://' + ip,
            'LDAPBASE'              : base,
            'BINDDN'                : binddn,
            'BINDPW'                : bindpw,
            'TIMEOUT'               : '5',
            'SIZELIMIT'             : '500',
            'LOGINNAMEATTRIBUTE'    : 'uid',
            'LDAPSEARCHFILTER'      : '(uid=*)(objectClass=inetOrgPerson)',
            'LDAPFILTER'            : '(&(uid=%s)(objectClass=inetOrgPerson))',
            'USERINFO'              : '{ "username": "uid", "phone" : "telephoneNumber", "mobile" : "mobile", "email" : "mail", "surname" : "sn", "givenname" : "givenName" }',
            'UIDTYPE'               : uidType,
            'NOREFERRALS'           : 'False'
        }
        response = self.app.get(url(controller='system', action='setResolver'), params=ldapDef)
        assert '"status": true,' in response

        ldapConf = self.getConfig(name, bindpw=bindpw, binddn=binddn)
        return ldapConf

    def getUserList4Resolver(self, resolver, username=None):

        param = {'username' : '*', 'resConf': resolver}
        if username is not None:
            param['username'] = username

        response = self.app.get(url(controller='admin', action='userlist'), params=param)
        if ("error") in response:
            body = json.loads(response.body)
            result = body.get('result')
            error = result.get('error')
            raise Exception(error.get('message'))
        else:
            assert ('"status": true,' in response)

        body = json.loads(response.body)
        result = body.get('result')
        userList = result.get('value')

        return userList

    def getUserList4Realm(self, realm, username=None):

        param = {'username' : '*', 'realm': realm}
        if username is not None:
            param['username'] = username

        response = self.app.get(url(controller='admin', action='userlist'), params=param)
        if ("error") in response:
            body = json.loads(response.body)
            result = body.get('result')
            error = result.get('error')
            raise Exception(error.get('message'))
        else:
            assert ('"status": true,' in response)

        body = json.loads(response.body)
        result = body.get('result')
        userList = result.get('value')

        return userList

    def check_ip(self, ip):
        ret = False

        return ret

    def getConfig(self, conf, bindpw=None, binddn=None):

        ret = {}

        response = self.app.get(url(controller='system', action='getConfig'))

        resp = json.loads(response.body)
        config = resp.get('result').get('value')

        for k in config:
            if k.startswith('ldapresolver.') and k.endswith(u'' + conf):
                    ret[k] = config.get(k)

        ''' finally fix the pw and dn in the config '''
        if bindpw is not None and len(ret) > 0:
            ret['ldapresolver.BINDPW.' + conf] = bindpw

        if binddn is not None and len(ret) > 0:
            ret['ldapresolver.BINDDN.' + conf] = binddn

        return ret

    def addToken(self, user, pin=None, serial=None, typ=None):

        if serial is None:
            serial = 's' + user
        if pin is None:
            pin = user
        if typ is None:
            typ = 'spass'
        param = { 'user': user, 'pin':pin, 'serial': serial, 'type':typ }
        response = self.app.get(url(controller='admin', action='init'), params=param)
        assert '"status": true,' in response

        return serial

    def authToken(self, user, passw=None):

        if passw is None:
            passw = user
        param = { 'user': user, 'pass':passw}
        response = self.app.get(url(controller='validate', action='check'), params=param)
        return response


    def showTokens(self, serial=None):

        param = {}
        if serial is not None:
            param['serial'] = serial
        response = self.app.get(url(controller='admin', action='show'), params=param)
        return response

    def delTokens(self, serial=None):

        param = {}
        if serial is not None:
            param['serial'] = serial
        response = self.app.get(url(controller='admin', action='remove'), params=param)
        return response


    def getLDAPUrl(self):

        self.ldapurl = None
        self.binddn = None
        self.bindpw = None
        self.ldapserverip = None

        #self.appconf = self.app.app.app.apps[1].application.app.application.app.app.app.config
        hostname = gethostname()

        if self.appconf.has_key('privacyidea.ldapTestServerIp.' + hostname):
            self.ldapurl = self.appconf.get('privacyidea.ldapTestServerIp.' + hostname)
        elif self.appconf.has_key('<include>') == True:
            try:
                filename = self.appconf.get('<include>')
                cfgParse = ConfigParser.ConfigParser()
                cfgParse.readfp(open(filename))
                incDict = cfgParse.defaults()
                self.ldapurl = incDict.get('privacyidea.ldapTestServerIp.'.lower() + hostname, None)
            except Exception as e:
                log.error('Error parsing include file: %r' % e)
        else:
                log.warning('no ldap Test server specified in the ini file!!')


        if self.ldapurl is not None:

            (part, _split, hostname) = self.ldapurl.partition('@')
            if part.startswith('ldap://'):  offstring = 'ldap://'
            if part.startswith('ldaps://'): offstring = 'ldaps://'

            userpass = part[len(offstring):]
            posCol = userpass.index(':')

            self.binddn = userpass[:posCol]
            self.bindpw = userpass[posCol + 1:]
            self.ldapserverip = hostname

        return (self.ldapurl, self.ldapserverip, self.binddn, self.bindpw)

    def setOtpPinPolicy(self, name='ldapOtpPin', realm='ldap_realm'):
        parameters = {
                         'name'     : name,
                         'user'     : '*',
                         'action'   : 'otppin=1, ',
                         'scope'    : 'authentication',
                         'realm'    : realm,
                         'time'     : '',
                         'client'   : '',
                         }
        response = self.app.get(url(controller='system', action='setPolicy'), params=parameters)
        log.error(response)

    def delOtpPinPolicy(self, name='ldapOtpPin'):
        parameters = { 'name' : name,
                      'selftest_admin' : 'superadmin' }
        response = self.app.get(url(controller='system', action='delPolicy'), params=parameters)
        log.debug(response)

    def test_one(self):
        '''
            LDAP test - test against a test LDAP server

        '''
        # get test server on a per test server host base
        (ldapurl, serverIp, binddn, bindpw) = self.getLDAPUrl()

        # Skip test if no config found
        if ldapurl is None or serverIp is None:
            skip_reason = "No ldap server test url like: ldap://cn=admin,dc=example,dc=com:test123!@192.168.0.2: defined in the *.ini file"
            if sys.version_info[0:2] >= (2, 7):
                # skipTest() has the advantage that it is shown in the test summary
                # but it is only available in Python 2.7
                self.skipTest(skip_reason)
            else:
                log.error("Skipping test 'test_one': " + skip_reason)
                return

        parameters = {'enableReplication' : 'true' }
        resp = self.app.get(url(controller='system', action='setConfig'), params=parameters)
        assert('"setConfig enableReplication:true": true' in resp)

        resolvers = []
        ldapObjects = []

        ldapServ = None

        ldapServerConnected = True

        ldapResolvers = [
            {'name' : 'ldapUsers', 'base': 'ou=users,dc=crypton,dc=info', 'uidType':'entryUUID' },
            {'name' : 'ldapPeople', 'base': 'ou=people,dc=crypton,dc=info', 'uidType':'entryUUID' },
        ]

        ''' add ldap server info to resolverdefinition '''
        for resolver in ldapResolvers:
            resolver['ip'] = serverIp
            resolver['bindpw'] = bindpw
            resolver['binddn'] = binddn

        thePassword = 'jörgi'
        try:
            ldapServ = LDAP(ldapurl)


            for entry in ldapResolvers:
                self.setupLDAPResolver(**entry)
                resolver = entry.get('name')
                resolvers.append(resolver)

                try:
                    base = entry.get('base')
                    path = base.split(',')
                    ou = path[0]
                    ou = ou[len('ou='):]
                    ldapServ.addOu(base, ou)
                    ldapObjects.append(base)

                except Exception as e:
                    if 'Already exists' not in unicode(e):
                        raise Exception(e)

                for user in self.users:
                    nuser = copy.deepcopy(user)
                    nuser['dn'] = user.get('dn') + entry.get('base')
                    nuser['uid'] = "%s-%s" % (user.get('uid'), resolver)
                    passw = hashlib.sha1(thePassword).digest()
                    passw = base64.encodestring(passw)
                    nuser['userPassword'] = '{SHA}%s' % (unicode(passw))

                    try:
                        ldapServ.addUser(**nuser)
                        ldapObjects.append(nuser.get('dn'))
                    except Exception as e:
                        if 'Already exists' not in unicode(e):
                            raise Exception(e)


        except Exception as e:
            log.error("%r" % e)
            log.error("%s" % traceback.format_exc())

            msg = unicode(e)
            if "Can't contact LDAP server" in msg:
                ldapServerConnected = False

        if ldapServerConnected == False:
            self.fail("No LDAP connection")


        realmresolvers = []
        tokenSerials = []
        try:
            for res in resolvers:
                realmresolvers.append('privacyidea.lib.resolvers.LDAPIdResolver.IdResolver.%s' % res)


            ''' next create a realm '''
            realmName = 'ldap_realm'
            parameters = {
                'realm'     : realmName,
                'resolvers' : u'%s' % (unicode(','.join(realmresolvers)))
            }
            resp = self.app.get(url(controller='system', action='setRealm'), params=parameters)
            assert('"value": true' in resp)

            resp = self.app.get(url(controller='system', action='getRealms'))
            assert('"default": "true"' in resp)

            ''' lookup for the user1 in realm '''
            userlist = self.getUserList4Realm(realmName, username='jörgi-*')
            assert len(userlist) == 2



            ''' now create and assign a token and validate '''
            for resolver in resolvers:
                userlist = self.getUserList4Resolver(resolver, username='jörgi-' + resolver)
                assert len(userlist) == 1
                user = 'jörgi-%s@%s' % (resolver, realmName)
                serial = self.addToken(user, pin='jörgi')
                ''' preserve the serials, so we can delete them later '''
                tokenSerials.append(serial)
                resp = self.authToken(user, passw='jörgi')
                assert '"value": true' in resp



            ''' change the ldap uidType from entryUid to DN'''
            for entry in ldapResolvers:
                entry['uidType'] = 'DN'
                self.setupLDAPResolver(**entry)

            ''' check that token is now an orphand one '''
            for serial in tokenSerials:
                resp = self.showTokens(serial=serial)
                assert '"User.username": "/:no user info:/"' in resp


            ''' token should be marked as orphand '''
            for resolver in resolvers:
                userlist = self.getUserList4Resolver(resolver, username='jörgi-' + resolver)
                assert len(userlist) == 1
                user = 'jörgi-%s@%s' % (resolver, realmName)

                resp = self.authToken(user, passw='jörgi')
                assert '"value": false' in resp

                ''' re-assing the token to the DN User1 '''
                self.addToken(user, pin='jörgi')
                resp = self.authToken(user, passw='jörgi')
                assert '"value": true' in resp

            '''set policy otppin=1 ant test checkPass with Umlaut'''
            self.setOtpPinPolicy(realm=realmName)
            for resolver in resolvers:
                user = 'jörgi-%s@%s' % (resolver, realmName)
                passw = thePassword
                resp = self.authToken(user, passw=passw)
                assert '"value": true' in resp


            ''' @todo: policy: passthrough '''


        except Exception as e:
            log.error(e)
            log.error("%s" % traceback.format_exc())
            raise Exception(e)

        finally:

            ''' cleanup : undefine realm and resolvers '''
            parameters = {"realm":realmName}
            resp = self.app.get(url(controller='system', action='delRealm'), params=parameters)
            assert('"result": true' in resp)

            for resolver in resolvers:
                parameters = {"resolver" : resolver}
                resp = self.app.get(url(controller='system', action='delResolver'), params=parameters)
                assert('"status": true' in resp)

            ''' check that token is now an orphand one '''
            for serial in tokenSerials:
                resp = self.showTokens(serial=serial)
                assert '"User.username": "/:no user info:/"' in resp
                resp = self.delTokens(serial=serial)


            if len(ldapObjects) > 0:
                ldapObjects.reverse()

            for ldapObject in ldapObjects:
                try:
                    ldapServ.delete(ldapObject)
                except Exception as e:
                    if 'No such object' not in unicode(e):
                        raise Exception(e)

            ldapServ.unbind()

            ''' 5 - cleanup'''
            parameters = {'key':'enableReplication' }
            resp = self.app.get(url(controller='system', action='delConfig'), params=parameters)
            assert('"delConfig enableReplication": true' in resp)

            self.delOtpPinPolicy()
        return







# -*- coding: utf-8 -*-
#
#  product:  privacyIDEA is a fork of LinOTP
#  module:   resolver library
#
#  May, 08 2014 Cornelius Kölbel
#  http://www.privacyidea.org
#
# 2014-10-03 fix getUsername function
#            Cornelius Kölbel <cornelius@privcyidea.org>
#
#  product:  LinOTP2
#  module:   useridresolver
#  tool:     UserIdResolver
#  edition:  Comunity Edition
#
#  Copyright (C) 2010 - 2014 LSE Leading Security Experts GmbH
#  License:  AGPLv3
#  contact:  http://www.linotp.org
#            http://www.lsexperts.de
#            linotp@lsexperts.de
#
# This code is free software; you can redistribute it and/or
# modify it under the terms of the GNU AFFERO GENERAL PUBLIC LICENSE
# License as published by the Free Software Foundation; either
# version 3 of the License, or any later version.
#
# This code is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU AFFERO GENERAL PUBLIC LICENSE for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
  Description:  This module implements the communication interface
                for resolvin user info to the user base

  Dependencies: -

UserIdResolver Interface class.

Defines the rough interface for a UserId Resolver
== a UserId Resolver is required to resolve the
   Login Name to an unique User Identifier

- for /etc/passwd this will be the uid
- for ldap this might be the DN
- for SQL the unique index ( what's the right name here (tm))

"""


class UserIdResolver(object):

    fields = {"username": 1, "userid": 1,
              "description": 0,
              "phone": 0, "mobile": 0, "email": 0,
              "givenname": 0, "surname": 0, "gender": 0
              }
    name = ""
    id = ""

    def __init(self):
        """
        init - usual bootstrap hook
        """
        self.name = "UserIdResolver"

    def close(self):
        """
        Hook to close down the resolver after one request
        """
        return

    @classmethod
    def getResolverClassType(cls):
        """
        provide the resolver type for registration
        """
        return 'UserIdResolver'

    def getResolverType(self):
        '''
        getResolverType - return the type of the resolver

        :return: returns the string 'ldapresolver'
        :rtype:  string
        '''
        return 'UserIdResolver'

    @classmethod
    def getResolverClassDescriptor(cls):
        '''
        return the descriptor of the resolver, which is
        - the class name and
        - the config description

        :return: resolver description dict
        :rtype:  dict
        '''
        descriptor = {}
        typ = cls.getResolverClassType()
        descriptor['clazz'] = "useridresolver.UserIdResolver"
        descriptor['config'] = {}
        return {typ: descriptor}

    def getResolverDescriptor(self):
        '''
        return the descriptor of the resolver, which is
        - the class name and
        - the config description

        :return: resolver description dict
        :rtype:  dict
        '''
        return UserIdResolver.getResolverClassDescriptor()

    def getUserId(self, loginName):
        """ getUserId(LoginName)
          - returns the identifier string
          - empty string if not exist

        """
        return self.id

    def getUsername(self, userid):
        """
        Returns the username/loginname for a given userid
        :param userid: The userid in this resolver
        :type userid: string
        :return: username
        :rtype: string
        """
        return self.name

    def getUserInfo(self, userid):
        """
        getUserInfo(UserID)
            This function returns all user information for a given user object
            identified by UserID.
        :return:  dictionary, if no object is found, the dictionary is empty
        """
        return ""

    def getUserList(self, serachDict):
        """
        This function finds the user objects,
        that have the term 'value' in the user object field 'key'

        :param searchDict:  dict with key values of user attributes -
                    the key may be something like 'loginname' or 'email'
                    the value is a regular expression.

        :return: list of dictionaries (each dictionary contains a
                 user object) or an empty string if no object is found.
        """
        return [{}]

    def getResolverId(self):
        """
        get resolver specific information
        :return: the resolver identifier string - empty string if not exist
        """
        return self.name

    def loadConfig(self, config, conf):
        return self

    def checkPass(self, uid, password):
        '''
        This function checks the password for a given uid.
        - returns true in case of success
        -         false if password does not match
        '''
        return False


def getResolverClass(packageName, className):
    """
    helper method to load the UserIdResolver class from a given
    package in literal. Checks, if the getUserId method exists,
    if not an error is thrown

    example:

        getResolverClass("PasswdIdResolver", "IdResolver")()

    :param packageName: the name package + module
    :param className: the name of the class, which should be loaded

    :return: the class object
    """
    mod = __import__(packageName, globals(), locals(), [className])
    klass = getattr(mod, className)
    ret = ""
    attribute = ""
    try:
        attrs = ["getUserId", "getUsername", "getUserInfo", "getUserList",
                 "checkPass", "loadConfig",
                 "getResolverId", "getResolverType", "getResolverDescriptor"
                 ]

        for att in attrs:
            attribute = att
            getattr(klass, att)
        ret = klass
    except:
        raise NameError("IdResolver AttributeError: " + packageName + "." +
                        className + " instance has no attribute '" +
                        attribute + "'")

    return ret



# -*- coding: utf-8 -*-
#
#  privacyIDEA
#  Jul 18, 2014 Cornelius Kölbel
#  License:  AGPLv3
#  contact:  http://www.privacyidea.org
#
# This code is free software; you can redistribute it and/or
# modify it under the terms of the GNU AFFERO GENERAL PUBLIC LICENSE
# License as published by the Free Software Foundation; either
# version 3 of the License, or any later version.
#
# This code is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU AFFERO GENERAL PUBLIC LICENSE for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
import logging
log = logging.getLogger(__name__)
from privacyidea.lib.log import log_with

try:
    from importlib import import_module as dyn_import
except:
    # importlib is not available in python 2.6
    def dyn_import(name):
        mod = __import__(name)
        components = name.split('.')
        for comp in components[1:]:
            mod = getattr(mod, comp)
        return mod


class MachineApplicationBase(object):

    application_name = "base"
    '''If bulk_call is false, the administrator may
    only retrieve authentication items for the
    very host he is starting the request.
    '''
    allow_bulk_call = False
        
    @classmethod
    def get_name(self):
        '''
        returns the identifying name of this application class
        '''
        return self.application_name
    
    def get_authentication_item(self,
                                token_type,
                                serial,
                                challenge=None):
        '''
        returns a dictionary of authentication items
        like public keys, challenges, responses...
        '''
        return "nothing"
    
    def get_options(self):
        '''
        returns a dictionary with a list of required and optional options
        '''
        return {'required': [],
                'optional': []}


@log_with(log)
def get_auth_item(application,
                  application_module,
                  token_type,
                  serial,
                  challenge=None):
    
    mod = dyn_import(application_module)
    auth_class = mod.MachineApplication()
    auth_item = auth_class.get_authentication_item(token_type,
                                                   serial,
                                                   challenge=challenge)
    return auth_item


@log_with(log)
def is_application_allow_bulk_call(application_module):
    mod = dyn_import(application_module)
    auth_class = mod.MachineApplication()
    return auth_class.allow_bulk_call



# -*- coding: utf-8 -*-
#
#  privacyIDEA is a fork of LinOTP
#  May 08, 2014 Cornelius Kölbel
#  License:  AGPLv3
#  contact:  http://www.privacyidea.org
#
#  Copyright (C) 2010 - 2014 LSE Leading Security Experts GmbH
#  License:  AGPLv3
#  contact:  http://www.linotp.org
#            http://www.lsexperts.de
#            linotp@lsexperts.de
#
# This code is free software; you can redistribute it and/or
# modify it under the terms of the GNU AFFERO GENERAL PUBLIC LICENSE
# License as published by the Free Software Foundation; either
# version 3 of the License, or any later version.
#
# This code is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU AFFERO GENERAL PUBLIC LICENSE for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#    
'''contains Errors and Exceptions
'''


import logging
log = logging.getLogger(__name__)


class privacyIDEAError(Exception):

    def __init__(self, description=u"privacyIDEAError!", id=10):
        self.id = id
        self.message = description
        Exception.__init__(self, description)

    def getId(self):
        return self.id

    def getDescription(self):
        return self.message

    def __unicode__(self):
        pstr = u"ERR%d: %r"
        if type(self.message) in [str, unicode]:
            pstr = u"ERR%d: %s"
        return pstr % (self.id, self.message)

    def __str__(self):
        pstr = u"ERR%d: %r"
        if type(self.message) in [str, unicode]:
            pstr = "ERR%d: %s"

        ### if we have here unicode, we might fail with conversion error
        try:
            res = pstr % (self.id, self.message)
        except Exception as exx:
            res = u"ERR%d: %r" % (self.id, self.message)
        return res


    def __repr__(self):
        ret = '%s(description=%r, id=%d)' % (type(self).__name__, self.message, self.id)
        return ret


class ValidateError(privacyIDEAError):
    def __init__(self, description="validation error!", id=10):
        privacyIDEAError.__init__(self, description=description, id=id)


class TokenAdminError(privacyIDEAError):
    def __init__(self, description="token admin error!", id=10):
        privacyIDEAError.__init__(self, description=description, id=id)


class ConfigAdminError(privacyIDEAError):
    def __init__(self, description="config admin error!", id=10):
        privacyIDEAError.__init__(self, description=description, id=id)


class UserError(privacyIDEAError):
    def __init__(self, description="user error!", id=905):
        privacyIDEAError.__init__(self, description=description, id=id)


class ServerError(privacyIDEAError):
    def __init__(self, description="server error!", id=905):
        privacyIDEAError.__init__(self, description=description, id=id)


class HSMException(privacyIDEAError):
    def __init__(self, description="hsm error!", id=707):
        privacyIDEAError.__init__(self, description=description, id=id)


class SelfserviceException(privacyIDEAError):
    def __init__(self, description="selfservice error!", id=807):
        privacyIDEAError.__init__(self, description=description, id=id)


class ParameterError(privacyIDEAError):
    def __init__(self, description="unspecified parameter error!", id=905):
        privacyIDEAError.__init__(self, description=description, id=id)


#eof###########################################################################




# -*- coding: utf-8 -*-
#  privacyIDEA is a fork of LinOTP
#  May 08, 2014 Cornelius Kölbel
#  License:  AGPLv3
#  contact:  http://www.privacyidea.org
#
#  Copyright (C) 2010 - 2014 LSE Leading Security Experts GmbH
#  License:  AGPLv3
#  contact:  http://www.linotp.org
#            http://www.lsexperts.de
#            linotp@lsexperts.de
#
# This code is free software; you can redistribute it and/or
# modify it under the terms of the GNU AFFERO GENERAL PUBLIC LICENSE
# License as published by the Free Software Foundation; either
# version 3 of the License, or any later version.
#
# This code is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU AFFERO GENERAL PUBLIC LICENSE for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
'''     
  Description:  functional tests
                
  Dependencies: -

'''
import datetime
import binascii
import hmac
import struct
import time
import random

import json

import traceback

from hashlib import sha1, sha256, sha512

from privacyidea.lib.crypto   import geturandom
from privacyidea.tests import TestController, url


import logging
log = logging.getLogger(__name__)


'''
  +-------------+--------------+------------------+----------+--------+
  |  Time (sec) |   UTC Time   | Value of T (hex) |   TOTP   |  Mode  |
  +-------------+--------------+------------------+----------+--------+
  |      59     |  1970-01-01  | 0000000000000001 | 94287082 |  SHA1  |
  |             |   00:00:59   |                  |          |        |
  |  1111111109 |  2005-03-18  | 00000000023523EC | 07081804 |  SHA1  |
  |             |   01:58:29   |                  |          |        |
  |  1111111111 |  2005-03-18  | 00000000023523ED | 14050471 |  SHA1  |
  |             |   01:58:31   |                  |          |        |
  |  1234567890 |  2009-02-13  | 000000000273EF07 | 89005924 |  SHA1  |
  |             |   23:31:30   |                  |          |        |
  |  2000000000 |  2033-05-18  | 0000000003F940AA | 69279037 |  SHA1  |
  |             |   03:33:20   |                  |          |        |
  | 20000000000 |  2603-10-11  | 0000000027BC86AA | 65353130 |  SHA1  |
  |             |   11:33:20   |                  |          |        |

  |      59     |  1970-01-01  | 0000000000000001 | 46119246 | SHA256 |
  |             |   00:00:59   |                  |          |        |
  |  1111111109 |  2005-03-18  | 00000000023523EC | 68084774 | SHA256 |
  |             |   01:58:29   |                  |          |        |
  |  1111111111 |  2005-03-18  | 00000000023523ED | 67062674 | SHA256 |
  |             |   01:58:31   |                  |          |        |
  |  1234567890 |  2009-02-13  | 000000000273EF07 | 91819424 | SHA256 |
  |             |   23:31:30   |                  |          |        |
  |  2000000000 |  2033-05-18  | 0000000003F940AA | 90698825 | SHA256 |
  |             |   03:33:20   |                  |          |        |
  | 20000000000 |  2603-10-11  | 0000000027BC86AA | 77737706 | SHA256 |
  |             |   11:33:20   |                  |          |        |

  |      59     |  1970-01-01  | 0000000000000001 | 90693936 | SHA512 |
  |             |   00:00:59   |                  |          |        |
  |  1111111109 |  2005-03-18  | 00000000023523EC | 25091201 | SHA512 |
  |             |   01:58:29   |                  |          |        |
  |  1111111111 |  2005-03-18  | 00000000023523ED | 99943326 | SHA512 |
  |             |   01:58:31   |                  |          |        |
  |  1234567890 |  2009-02-13  | 000000000273EF07 | 93441116 | SHA512 |
  |             |   23:31:30   |                  |          |        |
  |  2000000000 |  2033-05-18  | 0000000003F940AA | 38618901 | SHA512 |
  |             |   03:33:20   |                  |          |        |
  | 20000000000 |  2603-10-11  | 0000000027BC86AA | 47863826 | SHA512 |
  |             |   11:33:20   |                  |          |        |
  +-------------+--------------+------------------+----------+--------+



'''
seed = "3132333435363738393031323334353637383930"
seed32 = "3132333435363738393031323334353637383930313233343536373839303132"
seed64 = "31323334353637383930313233343536373839303132333435363738393031323334353637383930313233343536373839303132333435363738393031323334"

testvector = [
     { 'key'   : seed, 'timeStep' : 30,
       'hash'  : sha1, 'shash': 'sha1',
       'otps'  : [ (59, '94287082', '1970-01-01 00:00:59'), (1111111109, '07081804', '2005-03-18 01:58:29'),
                   (1111111111, '14050471', '2005-03-18 01:58:31'), (1234567890, '89005924', '2009-02-13 23:31:30'),
                   (2000000000, '69279037', '2033-05-18 03:33:20'), (20000000000, '65353130', '2603-10-11 11:33:20'), ] },

     { 'key'   : seed32, 'timeStep' : 30,
       'hash'  : sha256, 'shash': 'sha256',
       'otps'  : [ (59, '46119246', '1970-01-01 00:00:59'), (1111111109, '68084774' , '2005-03-18 01:58:29'),
                   (1111111111, '67062674', '2005-03-18 01:58:31'), (1234567890, '91819424' , '2009-02-13 23:31:30'),
                   (2000000000, '90698825', '2033-05-18 03:33:20'), (20000000000, '77737706', '2603-10-11 11:33:20'), ] },

     { 'key'   : seed64, 'timeStep' : 30,
       'hash'  : sha512, 'shash': 'sha512',
       'otps'  : [(59, '90693936', '1970-01-01 00:00:59'), (1111111109, '25091201', '2005-03-18 01:58:29'),
                  (1111111111, '99943326', '2005-03-18 01:58:31'), (1234567890, '93441116', '2009-02-13 23:31:30'),
                  (2000000000, '38618901', '2033-05-18 03:33:20'), (20000000000, '47863826', '2603-10-11 11:33:20'), ] }
]


class HmacOtp:

    def __init__(self, key, counter=0, digits=6, hashfunc=sha1):
        self.key = key
        self.counter = counter
        self.digits = digits
        self.hashfunc = hashfunc

    def hmac(self, key=None, counter=None):
        key = key or self.key
        counter = counter or self.counter
        digest = hmac.new(key, struct.pack(">Q", counter), self.hashfunc)
        return digest.digest()

    def truncate(self, digest):
        offset = ord(digest[-1:]) & 0x0f

        binary = (ord(digest[offset + 0]) & 0x7f) << 24
        binary |= (ord(digest[offset + 1]) & 0xff) << 16
        binary |= (ord(digest[offset + 2]) & 0xff) << 8
        binary |= (ord(digest[offset + 3]) & 0xff)

        return binary % (10 ** self.digits)

    def generate(self, key=None, counter=None):
        key = key or self.key
        counter = counter or self.counter
        otp = unicode(self.truncate(self.hmac(key, counter)))
        sotp = (self.digits - len(otp)) * "0" + otp
        return sotp

class TotpToken(object):

    def __init__(self, key=None, keylen=20, algo=None, digits=6, offset=0, jitter=0, timestep=60):


        ''' no key given - create one '''
        if key is None:
            self.key = binascii.hexlify(geturandom(keylen))
        else:
            self.key = key.decode('hex')
            keylen = len(self.key)

        if algo is None:
            if keylen == 20:
                algo = sha1
            elif keylen == 32:
                algo = sha256
            elif keylen == 64:
                algo = sha512
        else:
            algo = algo

        self.offset = offset
        self.jitter = jitter
        self.timestep = timestep
        self.digits = digits


        self.hmacOtp = HmacOtp(self.key, digits=self.digits, hashfunc=algo)

        return

    def getOtp(self, counter= -1, offset=0, jitter=0):
        '''
            @note: we require the ability to set the counter directly
                to validate the hmac token against the defined test vectors
        '''
        if counter == -1:

            if self.jitter != 0 or jitter != 0:
                jitter = random.randrange(-self.jitter , self.jitter)
            else:
                jitter = 0

            offset = self.offset + offset
            T0 = time.time() + offset + jitter
            counter = int((T0 / self.timestep) + 0.5)
        else:
            counter = int((counter / self.timestep) + 0.5)

        otp = self.hmacOtp.generate(counter=counter)

        return (otp, counter)

    def getKey(self):
        return self.key

    def getTimeStep(self):
        return self.timestep


    def getTimeFromCounter(self, counter):
        try:
            idate = int(counter - 0.5) * self.timestep
            ddate = datetime.datetime.utcfromtimestamp(idate / 1.0)
        except Exception as e:
            print "%r" % e
        return ddate



class TestTotpController(TestController):
    '''
    '''
    def setUp(self):
        TestController.setUp(self)
        self.serials = []



    def removeTokens(self):
        for serial in self.serials:
            self.delToken(serial)
        return

    def delToken(self, serial):
        p = {"serial" : serial }
        response = self.app.get(url(controller='admin', action='remove'), params=p)
        return response

    def time2float(self, curTime):
        '''
            time2float - convert a datetime object or an datetime sting into a float

            http://bugs.python.org/issue12750
        '''
        dt = datetime.datetime.now()
        if type(curTime) == datetime.datetime:
            dt = curTime
        elif type(curTime) == unicode:
            if '.' in curTime:
                tFormat = "%Y-%m-%d %H:%M:%S.%f"
            else:
                tFormat = "%Y-%m-%d %H:%M:%S"
            try:
                dt = datetime.datetime.strptime(curTime, tFormat)
            except Exception as e:
                log.error('[time2float] Error during conversion of datetime: %r' % e)
                log.error("[time2float] %s" % traceback.format_exc())
                raise Exception(e)
        else:
            log.error("[time2float] invalid curTime: %s. You need to specify a datetime.datetime" % type(curTime))
            raise Exception("[time2float] invalid curTime: %s. You need to specify a datetime.datetime" % type(curTime))

        td = (dt - datetime.datetime(1970, 1, 1))
        tCounter = (td.microseconds * 1.0 + (td.seconds + td.days * 24 * 3600) * 10 ** 6) / 10.0 ** 6

        return tCounter

    def addToken(self, user, pin=None, serial=None, typ=None, key=None, timeStep=60, timeShift=0, hashlib='sha1', otplen=8):

        if serial is None:
            serial = 's' + user

        if pin is None:
            pin = user

        if typ is None:
            typ = 'totp'

        param = { 'user': user, 'pin':pin, 'serial': serial, 'type':typ, 'timeStep':timeStep, 'otplen' : otplen, 'hashlib':hashlib}

        if key is not None:
            param['otpkey'] = key

        response = self.app.get(url(controller='admin', action='init'), params=param)
        assert '"status": true,' in response

        return serial

    def createTOTPToken(self, serial, seed):
        '''
            creates the test tokens
        '''
        parameters = {
                          "serial"  : serial,
                          "type"    : "TOTP",
                          # 64 byte key
                          "otpkey"  : seed,
                          "otppin"  : "1234",
                          "pin"     : "pin",
                          "otplen"  : 8,
                          "description" : "TOTP testtoken",
                          }

        response = self.app.get(url(controller='admin', action='init'), params=parameters)
        assert '"value": true' in response
        return


    def getTokenInfo(self, serial):
        param = { 'serial': serial}
        response = self.app.get(url(controller='admin', action='show'), params=param)
        assert '"status": true,' in response
        return json.loads(response.body)


    def checkOtp(self, user, otp, pin=None):
        if pin is None:
            pin = user

        param = { 'user': user, 'pass':pin + otp }
        response = self.app.get(url(controller='validate', action='check'), params=param)
        assert '"status": true,' in response

        return response


    def test_algo(self):
        '''
            totp test: verify that the local totp algorith is correct - selftest against testvector spec
        '''
        for tokData in testvector:
            key = tokData.get('key')
            algo = tokData.get('hash')
            step = tokData.get('timeStep')

            t1 = TotpToken(key, digits=8, algo=algo, timestep=step)
            otps = tokData.get('otps')
            for o in otps:
                counter = o[0]
                otp = o[1]
                if 20000000000 != counter:
                    tt = t1.getTimeFromCounter(counter)
                    log.debug("tokentime: %r" % tt)

                (rotp, _count) = t1.getOtp(counter=counter)
                if otp != rotp:
                    (rotp, _count) = t1.getOtp(counter=counter)
                    assert otp == rotp

        return

    def test_increment_timeshift(self):
        '''
            totp test: increments the time offset and verify the timeshift increases
        '''
        tokData = testvector[0]
        user = 'root'


        key = tokData.get('key')
        algo = tokData.get('hash')
        salgo = tokData.get('shash')
        step = tokData.get('timeStep')

        t1 = TotpToken(key, digits=8, algo=algo, timestep=step)
        step = t1.getTimeStep()

        tserial = self.addToken(user=user, typ='totp', key=key, timeStep=step, hashlib=salgo)
        self.serials.append(tserial)

        otpSet = set()

        for i in range(1, 5):
            offset = i * step
            (otp, counter) = t1.getOtp(offset=offset)
            tt = t1.getTimeFromCounter(counter)
            log.debug("tokentime: %r" % tt)

            res = self.checkOtp(user, otp)

            if otp not in otpSet:
                assert '"value": true' in res.body
                resInfo = self.getTokenInfo(tserial)
                tInfo = json.loads(resInfo.get('result').get('value').get('data')[0].get('privacyIDEA.TokenInfo'))
                tShift = tInfo.get('timeShift')
                assert tShift <= offset and tShift >= offset - step
            else:
                assert '"value": false' in res.body

            otpSet.add(otp)
            time.sleep(step / 2)
            log.debug('res')

        self.delToken('s'+user)
        return

    def test_decrement_timeshift(self):
        '''
            totp test: decrements the time offset and verify the timeshift decreases
        '''
        tokData = testvector[0]
        user = 'root'


        key = tokData.get('key')
        algo = tokData.get('hash')
        salgo = tokData.get('shash')
        step = tokData.get('timeStep')

        t1 = TotpToken(key, digits=8, algo=algo, timestep=step)
        step = t1.getTimeStep()

        self.delToken('s'+user)
        tserial = self.addToken(user=user, typ='totp', key=key, timeStep=step, hashlib=salgo)
        self.serials.append(tserial)

        otpSet = set()

        for i in range(1):
            offset = i * step * -1
            (otp, counter) = t1.getOtp(offset=offset)
            tt = t1.getTimeFromCounter(counter)
            log.debug("tokentime: %r" % tt)

            res = self.checkOtp(user, otp)

            if otp not in otpSet:
                if '"value": true' not in res.body:
                    assert '"value": true' in res.body
                resInfo = self.getTokenInfo(tserial)
                tInfo = json.loads(resInfo.get('result').get('value').get('data')[0].get('privacyIDEA.TokenInfo'))
                tShift = tInfo.get('timeShift')
                assert tShift <= offset and tShift >= offset - step
            else:
                assert '"value": false' in res.body

            otpSet.add(otp)
            time.sleep(step)
            log.debug('res')

        self.delToken('s'+user)
        return

    def test_use_token_twice(self):
        '''
            totp test: test if an otp could be used twice
        '''
        user = 'root'
        step = 30

        t1 = TotpToken(timestep=step)
        key = t1.getKey().encode('hex')
        step = t1.getTimeStep()

        tserial = self.addToken(user=user, otplen=t1.digits, typ='totp', key=key, timeStep=step)

        self.serials.append(tserial)

        (otp, counter) = t1.getOtp()
        tt = t1.getTimeFromCounter(counter)
        log.debug("tokentime: %r" % tt)

        res = self.checkOtp(user, otp)
        assert '"value": true' in res.body

        ''' reusing the otp again will fail'''
        res = self.checkOtp(user, otp)
        assert '"value": false' in res.body

        time.sleep(step)


        ''' after a while, we could do a check again'''
        (otp, counter) = t1.getOtp()
        tt = t1.getTimeFromCounter(counter)
        log.debug("tokentime: %r" % tt)

        res = self.checkOtp(user, otp)
        assert '"value": true' in res.body

        return


    def test_getotp(self):
        '''
            totp test: test the getotp - verify that in the list of getotp is the correct start otp of our test vector
        '''


        parameters = { 'name' : 'getmultitoken',
                       'scope' : 'gettoken',
                       'realm' : 'mydefrealm',
                       'action' : 'max_count_dpw=10, max_count_hotp=10, max_count_totp=10',
                       'user' : 'admin'
                      }
        response = self.app.get(url(controller='system', action='setPolicy'), params=parameters)


        tFormat = "%Y-%m-%d %H:%M:%S"

        for tokData in testvector:
            tkey = tokData.get('key')
            algo = tokData.get('hash')
            salgo = tokData.get('shash')
            step = tokData.get('timeStep')

            t1 = TotpToken(key=tkey, digits=8, algo=algo, timestep=step)
            tserial = self.addToken(user='root', typ='totp', key=tkey, timeStep=step, hashlib=salgo)

            self.serials.append(tserial)

            otps = tokData.get('otps')
            for o in otps:
                tCounter = o[0]
                counter = int(((tCounter) / step))
                otp = o[1]
                curTime = o[2]

                (cotp, ccounter) = t1.getOtp(counter=tCounter)
                tt = t1.getTimeFromCounter(ccounter)
                log.debug("tokentime: %r" % tt)

                parameters = {'serial' : tserial,
                              'curTime' : curTime,
                              'count' : "20",
                              'selftest_admin' : 'admin' }
                response = self.app.get(url(controller='gettoken', action='getmultiotp'), params=parameters)

                print response
                resp = json.loads(response.body)

                otpres = resp.get('result').get('value').get('otp')

                otp1 = otpres.get(str(counter))

                ''' calculate the utc offset'''
                uTime = otp1.get('time')
                fTime = self.time2float(uTime)
                ut = datetime.datetime.utcfromtimestamp(fTime)
                ot = datetime.datetime.fromtimestamp(fTime)
                td = (ut - ot)
                delta = (td.microseconds * 1.0 + (td.seconds + td.days * 24 * 3600) * 10 ** 6) / 10.0 ** 6

                ''' compare the time absolute values'''
                dt = datetime.datetime.strptime(uTime, tFormat) + datetime.timedelta(seconds=delta)
                ct = datetime.datetime.strptime(curTime, tFormat)
                cl = ct - datetime.timedelta(seconds=step)

                assert dt <= ct and cl <= dt
                assert otp1.get('otpval') == otp


            self.removeTokens()

        return



    def tearDown(self):
        self.removeTokens()

        TestController.tearDown(self)



# -*- coding: utf-8 -*-
#  privacyIDEA is a fork of LinOTP
#  May 08, 2014 Cornelius Kölbel
#  License:  AGPLv3
#  contact:  http://www.privacyidea.org
#
#  Copyright (C) 2010 - 2014 LSE Leading Security Experts GmbH
#  License:  AGPLv3
#  contact:  http://www.linotp.org
#            http://www.lsexperts.de
#            linotp@lsexperts.de
#
# This code is free software; you can redistribute it and/or
# modify it under the terms of the GNU AFFERO GENERAL PUBLIC LICENSE
# License as published by the Free Software Foundation; either
# version 3 of the License, or any later version.
#
# This code is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU AFFERO GENERAL PUBLIC LICENSE for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
'''
request and reply handling
'''

import qrcode
import StringIO
import urllib

import json

from pylons import request
from pylons import tmpl_context as c

from privacyidea.lib.error       import privacyIDEAError
from privacyidea.lib.util        import get_version
from privacyidea.lib.log import log_with

optional = True
required = False

PRIVACYIDEA_ERRORS = [707]

httpErr = {
        '400': 'Bad Request',
        '401': 'Unauthorized',
        '403': 'Forbidden',
        '404': 'Not Found',
        '410': 'Gone',
        '500': 'Internal Server Error',
        '501': 'Not Implemented',
        '502': 'Bad Gateway',
        '503': 'Service Unavailable',
        }

resp = """
<html>
<head>
<title>%s %s</title>
</head>
<body>
<h1>%s %s</h1>
%s
<br>
<br>
</body>
</html>
"""

import logging
log = logging.getLogger(__name__)

def sendError(response, exception, id=1, context=None):
    '''
    sendError - return a json error result document

    remark:
     the 'context' is especially required to catch errors from the _before_
     methods. The return of a _before_ must be of type response and
     must have the attribute response._exception set, to stop further
     processing, which otherwise will have ugly results!!

    :param response:  the pylon response object
    :type  response:  response object
    :param exception: should be a privacyidea exception (s. privacyidea.lib.error.py)
    :type  exception: exception
    :param id:        id value, for future versions
    :type  id:        int
    :param context:   default is None or 'before'
    :type  context:   string

    :return:     json rendered sting result
    :rtype:      string

    '''
    ret = ''
    errId = -311

    ## handle the different types of exception:
    ## Exception, privacyIDEAError, str/unicode
    if hasattr(exception, '__class__') == True \
    and isinstance(exception, Exception):
        errDesc = unicode(exception)
        if isinstance(exception, privacyIDEAError):
            errId = exception.getId()

    elif type(exception) in [str, unicode]:
        errDesc = unicode(exception)

    else:
        errDesc = u"%r" % exception

    HTTP_ERROR = False
    ## check if we have an additional request parameter 'httperror'
    ## which triggers the error to be delivered as HTTP Error
    try:
        httperror = request.params.get('httperror')
    except Exception as exx:
        httperror = "%r" % exx

    if httperror is not None:
        ## now lookup in the config, which privacyidea errors should be shwon as
        ## HTTP error
        privacyidea_errors = c.privacyideaConfig.get('privacyidea.errors', None)
        if privacyidea_errors is None:
            HTTP_ERROR = True
        else:
            privacyidea_errors = privacyidea_errors.split(',')
            if unicode(errId) in privacyidea_errors:
                HTTP_ERROR = True
            else:
                HTTP_ERROR = False

    if HTTP_ERROR is True:
        ## httperror as param exist but is not defined
        ## so fallback to 500 - Internal Server Error
        if len(httperror) == 0: httperror = '500'

        ## prepare the response to be of text/html
        response.content_type = 'text/html'
        response.status = httperror

        code = httperror
        status = httpErr.get(httperror, '')
        desc = '[%s] %d: %s' % (get_version(), errId, errDesc)
        ret = resp % (code, status, code, status, desc)

        if context in ['before', 'after']:
            response._exception = exception
            response.text = u'' + ret
            ret = response

    else:
        response.content_type = 'application/json'
        res = { "jsonrpc": "2.0",
                "result" :
                    {"status": False,
                        "error": {
                            "code"    :   errId,
                            "message" :   errDesc,
                            },
                    },
                 "version": get_version(),
                 "id": id
            }

        ret = json.dumps(res, indent=3)

        if context in ['before', 'after']:
            response._exception = exception
            response.body = ret
            ret = response

    return ret


def sendResult(response, obj, id=1, opt=None):
    '''
        sendResult - return an json result document

        :param response: the pylons response object
        :type  response: response object
        :param obj:      simple result object like dict, sting or list
        :type  obj:      dict or list or string/unicode
        :param  id:      id value, for future versions
        :type   id:      int
        :param opt:      optional parameter, which allows to provide more detail
        :type  opt:      None or simple type like dict, list or string/unicode

        :return:     json rendered sting result
        :rtype:      string

    '''

    response.content_type = 'application/json'

    res = { "jsonrpc": "2.0",
            "result": { "status": True,
                        "value": obj,
                      },
           "version": get_version(),
           "id": id }

    if opt is not None and len(opt) > 0:
        res["detail"] = opt

    return json.dumps(res, indent=3)

def sendCSVResult(response, obj, flat_lines=False, filename="privacyidea-tokendata.csv"):
    '''
    returns a CSV document of the input data (like in /admin/show)

    :param response: The pylons response object
    :param obj: The data, that gets serialized as CSV
    :type obj: JSON object
    :param flat_lines: If True the object only contains a list of the dict { 'cell': ..., 'id': ... } 
                        as in all the flexigrid functions.
    'type flat_lines: boolean
    '''
    delim = "'"
    response.content_type = "application/force-download"
    response.headers['Content-disposition'] = 'attachment; filename=%s' % filename
    output = u""

    if not flat_lines:
        # Do the header
        for k, v in obj.get("data", {})[0].iteritems():
            output += "%s%s%s, " % (delim, k, delim)
        output += "\n"

        # Do the data
        for row in obj.get("data", {}):
            for val in row.values():
                if type(val) in [str, unicode]:
                    value = val.replace("\n", " ")
                else:
                    value = val
                output += "%s%s%s, " % (delim, value, delim)
            output += "\n"
    else:
        for l in obj:
            for elem in l.get("cell", []):
                output += "'%s', " % elem
            
            output += "\n"

    return output

def sendXMLResult(response, obj, id=1):
    response.content_type = 'text/xml'
    res = '<?xml version="1.0" encoding="UTF-8"?>\
            <jsonrpc version="2.0">\
            <result>\
                <status>True</status>\
                <value>%s</value>\
            </result>\
            <version>%s</version>\
            <id>%s</id>\
            </jsonrpc>' % (obj, get_version(), id)
    return res


def sendXMLError(response, exception, id=1):
    response.content_type = 'text/xml'
    if not hasattr(exception, "getId"):
        errId = -311
        errDesc = str(exception)
    else:
        errId = exception.getId()
        errDesc = exception.getDescription()
    res = '<?xml version="1.0" encoding="UTF-8"?>\
            <jsonrpc version="2.0">\
            <result>\
                <status>False</status>\
                <error>\
                    <code>%s</code>\
                    <message>%s</message>\
                </error>\
            </result>\
            <version>%s</version>\
            <id>%s</id>\
            </jsonrpc>' % (errId, errDesc, get_version(), id)
    return res

@log_with(log)
def sendQRImageResult(response, data, param=None, id=1, typ='html'):
    '''
    method
        sendQRImageResult

    arguments
        response - the pylon response object
        param    - the paramters of the request
        id       -
        html     - print qrcode wrapped by html or not

    '''
    width = 0
    alt = None
    ret = None

    if param is None:
        param = {}

    if 'qr' in param:
        typ = param.get('qr')
        del param['qr']

    if 'width' in param:
        width = param.get('width')
        del param['width']

    if 'alt' in param:
        alt = param.get('alt')
        del param['alt']

    if typ in ['img', 'embed']:
        response.content_type = 'text/html'
        ret = create_img(data, width, alt)

    elif typ in ['png']:
        response.content_type = 'image/png'
        ret = create_png(data)
        response.content_length = len(ret)

    else:
        response.content_type = 'text/html'
        ret = create_html(data, width, param)

    return ret


def create_png(data, alt=None):
    '''

    '''

    img = qrcode.make(data)

    output = StringIO.StringIO()
    img.save(output)
    o_data = output.getvalue()
    output.close()

    return o_data


def create_img(data, width=0, alt=None):
    '''
        _create_img - create the qr image data

        :param data: input data that will be munched into the qrcode
        :type  data: string
        :param width: image width in pixel
        :type  width: int

        :return: <img/> taged data
        :rtype:  string
    '''
    width_str = ''
    alt_str = ''

    o_data = create_png(data, alt=alt)
    data_uri = o_data.encode("base64").replace("\n", "")

    if width != 0:
        width_str = " width=%d " % (int(width))

    if alt is not None:
        val = urllib.urlencode({'alt':alt})
        alt_str = " alt=%r " % (val[len('alt='):])

    ret_img = '<img %s  %s  src="data:image/png;base64,%s"/>' % (alt_str, width_str, data_uri)

    return ret_img


def create_html(data, width=0, alt=None):
    '''
        _create_html - create the qr image data embeded in html tag

        :param data: input data that will be munched into the qrcode
        :type  data: string
        :param width: image width in pixel
        :type  width: int

        :return: <img/> taged data
        :rtype:  string
    '''
    alt_str = ''

    img = create_img(data, width=width, alt=alt)
    if alt is not None:
        if type(alt) in (str, u''):
            alt_str = '<p>%s</p>' % alt
        elif type(alt) == dict:
            alta = []
            for k in alt.keys():
                alta.append('<li> %s:%s </li>' % (k, alt.get(k)))
            alt_str = '<ul>%s</ul>' % " ".join(alta)
        elif type(alt) == list:
            alta = []
            for k in alt:
                alta.append('<li> %s </li>' % (k))


    ret_html = '<html><body>%s%s</body></html>' % (img , alt_str)

    return ret_html


#eof#######################################################




# -*- coding: utf-8 -*-
#
#  privacyIDEA is a fork of LinOTP
#  May 08, 2014 Cornelius Kölbel
#  License:  AGPLv3
#  contact:  http://www.privacyidea.org
#
#  Copyright (C) 2010 - 2014 LSE Leading Security Experts GmbH
#  License:  AGPLv3
#  contact:  http://www.linotp.org
#            http://www.lsexperts.de
#            linotp@lsexperts.de
#
# This code is free software; you can redistribute it and/or
# modify it under the terms of the GNU AFFERO GENERAL PUBLIC LICENSE
# License as published by the Free Software Foundation; either
# version 3 of the License, or any later version.
#
# This code is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU AFFERO GENERAL PUBLIC LICENSE for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
'''            
  Description:  base functions for the selfservice

  Dependencies: -

'''


from pylons import config
from privacyidea.lib.log import log_with
import logging
log = logging.getLogger(__name__)

@log_with(log)
def get_imprint(realm):
    '''
    This function returns the imprint for a certai realm.
    This is just the contents of the file <realm>.imprint in the directory
    <imprint_directory>
    '''
    res = ""
    realm = realm.lower()
    directory = config.get("privacyidea.imprint_directory", "/etc/privacyidea/imprint")
    filename = "%s/%s.imprint" % (directory, realm)
    try:
        pass
        f = open(filename)
        res = f.read()
        f.close()
    except Exception as e:
        log.info("can not read imprint file: %s. (%r)"
                 % (filename, e))

    return res



'''
privacyIDEA is an open solution for strong two-factor authentication with One Time Passwords.

The core modules and basic necessary components are licensed under the AGPLv3, so that you 
are able to have a complete working open source solution. But privacyIDEA is also open as far 
as its modular architecture is concerned.   
privacyIDEA aims to not bind you to any  decision of the authentication protocol or it does not 
dictate you where your user information should be stored. This is achieved by its new, totally 
modular architecture. 

This is the controller module. The controllers provide the Web API to communicate with privacyIDEA.

You can use the following controllers:

account		- used for loggin in to the selfservice
admin		- API to manage the tokens
audit		- to search the audit trail
auth		- to do authentication tests
error		- to display errors
gettoken	- to retrieve OTP values
license		- to manage license
manage		- the Web UI
openid		- the openid interface
selfservice	- the selfservice UI
system		- to configure the system
testing		- for testing purposes
validate	- for authenticating/ OTP checking


'''




# -*- coding: utf-8 -*-
#  privacyIDEA is a fork of LinOTP
#  May 08, 2014 Cornelius Kölbel
#  License:  AGPLv3
#  contact:  http://www.privacyidea.org
#
#  Copyright (C) 2010 - 2014 LSE Leading Security Experts GmbH
#  License:  AGPLv3
#  contact:  http://www.linotp.org
#            http://www.lsexperts.de
#            linotp@lsexperts.de
#
# This code is free software; you can redistribute it and/or
# modify it under the terms of the GNU AFFERO GENERAL PUBLIC LICENSE
# License as published by the Free Software Foundation; either
# version 3 of the License, or any later version.
#
# This code is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU AFFERO GENERAL PUBLIC LICENSE for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
'''            
  Description:  functional tests
                
  Dependencies: -

'''
import binascii

import urllib
import httplib2
import json

from privacyidea.tests import TestController, url
from privacyidea.lib.HMAC import HmacOtp
from mock import patch
import smtplib
import re
import time

SMS=True

import privacyidea.smsprovider.HttpSMSProvider


import logging
log = logging.getLogger(__name__)
# A reduced TOTP timestep to speed up tests
TOTP_TIMESTEP=30

def email_otp_func(call_args):
    '''
    callback to extract the otp value from the mock interface parameters

    :param call_args: arguments to the smtp.SMTP.sendmail method
    :return: the extracted otp value as string
    '''
    otp = None
    try:
        ordered_args = call_args[0]
        _email_from = ordered_args[0]
        _email_to = ordered_args[1]
        message = ordered_args[2]
        matches = re.search('\d{6}', message)
        otp = matches.group(0)
    except Exception as _exx:
        pass
    return otp

def sms_otp_func(call_args):
    '''
    callback to extract the otp value from the mock interface parameters

    :param call_args: arguments to the smtp.SMTP.sendmail method
    :return: the extracted otp value as string
    '''
    otp = None
    try:
        ordered_args = call_args[0]
        _phone = ordered_args[0]
        otp = ordered_args[1]
    except Exception as _exx:
        pass
    return otp


def get_otp(counter=0, otpkey=None, mock_obj=None, otp_func=None, typ='hmac'):
    '''
    extract from the context the otp value
    - if we have a mock_obj and a extractor callback, we are using this one
    - else we take the given otp value and the secret to calculate the new one

    :param counter: counter base for the otp calculation
    :param otpkey: the otpkey secret
    :param mock_obj: the mock hooked function which recieved the parameters
    :param otp_func: the otp extractor function
    '''
    otp = None
    if mock_obj is not None:
        call_args = mock_obj.call_args
        ## compare type of otp_func with known function
        if otp_func is not None and type(otp_func) == type(get_otp):
            otp = otp_func(call_args)

    if otp is None:
        counter = counter + 1
        otp = calcOTP(otpkey, counter=counter, typ=typ)

    return otp

def calcOTP(key, counter=0, digits=6, typ=None):
    '''
    as we have to use this method in a not class related function
    this function is extracted

    :param key: the otpkey secret
    :param counter: the related counter
    :param digits: the number of to be returned digits

    :return: the otp value as string
    '''
    htoken = HmacOtp(digits=digits)
    if typ == 'totp':
        counter = int((time.time() / TOTP_TIMESTEP) + 0.5)

    otp = htoken.generate(counter=counter, key=binascii.unhexlify(key))

    return otp


def do_http(remoteServer, params=None):

        if params is None:
            params = {}

        data = urllib.urlencode(params)
        request_url = "%s" % (remoteServer)
        content = None

        try:
            ## prepare the submit and receive headers
            headers = {"Content-type": "application/x-www-form-urlencoded",
                       "Accept": "text/plain"}

            ## submit the request
            http = httplib2.Http()
            (resp, content) = http.request(request_url,
                                           method="POST",
                                           body=data,
                                           headers=headers)
            print "requesturl: ", request_url
            print "resp:", resp
        except Exception as exx:
            log.error("Error %r" % exx)

        return content

class TestChallengeResponseController(TestController):

    def setUp(self):
        '''
        This sets up all the resolvers and realms
        '''
        TestController.setUp(self)

        if hasattr(self, "policies") is False:
            setattr(self, "policies", [])

        if hasattr(self, "serials") is False:
            setattr(self, "serials", [])

        self.patch_smtp = None
        self.patch_sms = None

        self.deleteAllTokens()
        self.deleteAllPolicies()

        return

    def tearDown(self):
        if self.patch_smtp is not None:
            self.patch_smtp.stop()
        if self.patch_sms is not None:
            self.patch_sms.stop()
        TestController.tearDown(self)

    def deleteAllPolicies(self):
        '''
        '''
        response = self.app.get(url(controller='system', action='getPolicy'),)
        self.assertTrue('"status": true' in response, response)

        body = json.loads(response.body)
        policies = body.get('result', {}).get('value', {}).keys()

        for policy in policies:
            self.delPolicy(policy)

        return

    def deleteAllTokens(self):
        serials = []

        response = self.app.get(url(controller='admin', action='show'),
                                )
        self.assertTrue('"status": true' in response, response)

        body = json.loads(response.body)
        tokens = body.get('result', {}).get('value', {}).get('data', {})
        for token in tokens:
            serial = token.get("privacyIDEA.TokenSerialnumber")
            serials.append(serial)

        for serial in serials:
            self.removeTokenBySerial(serial)

        return

    def calcOTP(self, key, counter=0, digits=6, typ='hmac'):
        otp = calcOTP(key, counter=counter, digits=digits, typ=typ)
        return otp

    def createToken(self, serial='F722362', user='root', pin="pin",
                    description="TestToken1", typ='hmac',
                    otpkey="AD8EABE235FC57C815B26CEF3709075580B44738",
                    phone=None,
                    email_address=None
                    ):

        parameters = {
                        "serial"  : serial,
                        "otpkey"  : otpkey,
                        "user"    : user,
                        "pin"     : pin,
                        "type"    : typ,
                        "description" : description,
                      }
        if phone is not None:
            parameters['phone'] = phone
        if email_address is not None:
            parameters['email_address'] = email_address

        response = self.app.get(url(controller='admin', action='init'),
                                params=parameters)
        self.assertTrue('"value": true' in response, response)
        self.serials.append(serial)
        
        if typ.lower() == "totp":
            # we do not use the default timestep!
            response = self.app.get(url(controller='admin', action='set'),
                                params={"serial" : serial,
                                        "timeStep" : TOTP_TIMESTEP})
            self.assertTrue('"set timeStep": 1' in response, response)
        return serial

    def removeTokenBySerial(self, serial):
        parameters = {
                      "serial": serial,
                      }

        response = self.app.get(url(controller='admin', action='remove'),
                                params=parameters)
        return response

    def setPinPolicy(self, name='otpPin', realm='ldap_realm',
                     action='otppin=1, ', scope='authentication',
                     active=True, remoteurl=None):
        parameters = {
                         'name'     : name,
                         'user'     : '*',
                         'action'   : action,
                         'scope'    : scope,
                         'realm'    : realm,
                         'time'     : '',
                         'client'   : '',
                         'active'   : active,
                         }

        r_url = url(controller='system', action='setPolicy')

        if remoteurl is not None:
            r_url = "%s/%s" % (remoteurl, "system/setPolicy")
            response = do_http(r_url, params=parameters)
        else:
            response = self.app.get(r_url, params=parameters)



        self.assertTrue('"status": true' in response, response)


        r_url = url(controller='system', action='getPolicy')

        if remoteurl is not None:
            r_url = "%s/%s" % (remoteurl, "system/getPolicy")
            response = do_http(r_url, params=parameters)
        else:
            response = self.app.get(r_url, params=parameters)

        self.assertTrue('"status": true' in response, response)


        self.policies.append(name)
        return response

    def delPolicy(self, name='otpPin', remoteurl=None):

        parameters = {'name': name,
                      'selftest_admin': 'superadmin'
                      }
        r_url = url(controller='system', action='delPolicy')

        if remoteurl is not None:
            r_url = "%s/%s" % (remoteurl, "system/delPolicy")
            response = do_http(r_url, params=parameters)
        else:
            response = self.app.get(r_url, params=parameters)


        return response



    def setup_remote_token(self,
                           typ="pw",
                           otpkey="123456",
                           remoteurl="http://127.0.0.1:5001"):
        # local token
        serials = []
        params = [
                  ## the token set with remote pin checking
                  {
                        "serial"  : "LSRE001",
                        "type"    : "remote",
                        "otpkey"  : otpkey,
                        "otppin"  : "",
                        "user"    : "remoteuser",
                        "pin"     : "lpin",
                        "description" : "RemoteToken1",
                        'remote.server' : remoteurl,
                        'remote.local_checkpin' : 0,
                        'remote.serial' : 'LSPW1',
                      },
                  { # the target is accessed via serial, so no user is required
                        "serial"  : "LSPW1",
                        "type"    : typ,
                        "otpkey"  : otpkey,
                        "otppin"  : "",
                        "user"    : "",
                        "pin"     : "rpin",
                  },
                  ## the token set with local pin checking
                  {
                        "serial"  : "LSRE002",
                        "type"    : "remote",
                        "otpkey"  : otpkey,
                        "user"    : "localuser",
                        "pin"     : "lpin",
                        "description" : "RemoteToken2",
                        'remote.server' : remoteurl,
                        'remote.local_checkpin' : 1,
                        'remote.serial' : 'LSPW2',
                        },
                  { # the target is accessed via serial, so no user is required
                        "serial"  : "LSPW2",
                        "type"    : typ,
                        "otpkey"  : otpkey,
                        "otppin"  : "",
                        "user"    : "",
                        "pin"     : "",
                         },
                  ]
        for param in params:
            serials.append(param.get('serial'))
            response = self.app.get(url(controller='admin', action='init'),
                                    params=param)
            self.assertTrue('"value": true' in response, response)

        ## enforce the awareness of policy changes
        parameters = {'enableReplication' : 'true' }
        resp = self.app.get(url(controller='system', action='setConfig'),
                            params=parameters)
        self.assertTrue('"setConfig enableReplication:true": true' in resp, resp)

        parameters = {'enableReplication' : 'true' }
        resp = do_http("%s/system/setConfig" % remoteurl,
                            params=parameters)
        print "remoteurl:", remoteurl
        print "parameters:", parameters
        print "Response:", resp
        self.assertTrue('"setConfig enableReplication:true": true' in resp, resp)

        return serials

    def setup_radius_token(self):

        serials = []

        # The token with the remote PIN
        parameters = [{
                      "serial"  : "radius1",
                      "type"    : "radius",
                      "otpkey"  : "1234567890123456",
                      "otppin"  : "",
                      "user"    : "remoteuser",
                      "pin"     : "",
                      "description" : "RadiusToken1",
                      'radius.server' : 'localhost:18012',
                      'radius.local_checkpin' : 0,
                      'radius.user' : 'challenge',
                      'radius.secret' : 'testing123',
                      },

                    # the token with the local PIN
                    {
                      "serial"  : "radius2",
                      "type"    : "radius",
                      "otpkey"  : "1234567890123456",
                      "otppin"  : "local",
                      "user"    : "localuser",
                      "pin"     : "local",
                      "description" : "RadiusToken2",
                      'radius.server' : 'localhost:18012',
                      'radius.local_checkpin' : 1,
                      'radius.user' : 'user_no_pin',
                      'radius.secret' : 'testing123',
                      },
                     ]
        for param in parameters:
            response = self.app.get(url(controller='admin', action='init'),
                                    params=param)
            self.assertTrue('"value": true' in response, response)
            serials.append(param.get("serial"))

        return serials

    def _start_radius_server(self):
        """
        Start the dummy radius server

        We need to start the radius server for every test, since every test instatiates a new TestClass and thus the
        radius server process will not be accessable outside of a test anymore
        """
        import subprocess
        try:
            self.p = subprocess.Popen(["../../tools/dummy_radius_server.py",
                                       "-d", "config/dictionary"])
        except Exception as exx:
            raise exx
        assert self.p is not None

    def _stop_radius_server(self):
        '''
        stopping the dummy radius server
        '''
        if self.p:
            r = self.p.kill()
            log.debug(r)

        return

    def test_03_hmac_regression(self):
        '''
        Challenge Response Test: test if HMAC tokens still work -
                this is a potential challenge token
        '''

        counter = 0
        otpkey = "AD8EABE235FC57C815B26CEF3709075580B44738"
        ## normal test
        serial = self.createToken(pin="shortpin", otpkey=otpkey,
                                                    user='passthru_user1')

        otp = self.calcOTP(otpkey, counter=counter)
        ## submit a pin only request - to trigger a challenge
        parameters = {"user": "passthru_user1", "pass": "shortpin" + otp}
        response = self.app.get(url(controller='validate', action='check'),
                                                            params=parameters)
        self.assertTrue('"value": true' in response, response)
        self.removeTokenBySerial(serial)


        ## with otppin==1 the pin should be the same
        self.setPinPolicy(realm='myDefRealm', action='otppin=1, ')

        serial = self.createToken(pin="otppin", user='passthru_user1')

        otp = self.calcOTP(otpkey, counter=counter)
        parameters = {"user": "passthru_user1", "pass": "geheim1" + otp}
        response = self.app.get(url(controller='validate', action='check'),
                                params=parameters)
        self.assertTrue('"value": true' in response, response)
        self.removeTokenBySerial(serial)

        self.delPolicy()

        ## with otppin==2 the pin is not required at all
        self.setPinPolicy(realm='myDefRealm', action='otppin=2, ')

        serial = self.createToken(pin="otppin", user='passthru_user1')
        otp = self.calcOTP(otpkey, counter=1)
        parameters = {"user": "passthru_user1", "pass": otp}
        response = self.app.get(url(controller='validate', action='check'),
                                params=parameters)
        self.assertTrue('"value": true' in response, response)

        ## finally otppin == 2 and wrong otp would trigger
        parameters = {"user": "passthru_user1", "pass": "123456"}
        response = self.app.get(url(controller='validate', action='check'),
                                params=parameters)
        if '"value": true' not in response:
            log.error(response)
        self.assertTrue('"value": false' in response, response)
        self.assertTrue('transactionid"' not in response, response)

        self.setPinPolicy(name="ch_resp", realm='myDefRealm',
                                            action='challenge_response=hmac, ')

        ## no challenge request - empty pin + otp does not match
        parameters = {"user": "passthru_user1", "pass": "123456"}
        response = self.app.get(url(controller='validate', action='check'),
                                params=parameters)
        if '"value": true' not in response:
            log.error(response)
        self.assertTrue('"value": false' in response, response)
        self.assertTrue('transactionid"' not in response, response)

        self.removeTokenBySerial(serial)
        self.delPolicy(name="ch_resp")

        return


    def test_02_spass_regression(self):
        '''
        Challenge Response Test: test if SPASS tokens still work - it is a no challenge token
        '''

        ## normal test
        serial = self.createToken(pin="shortpin", typ='spass',
                                                        user='passthru_user1')

        ## submit a pin only request - to trigger a challenge
        parameters = {"user": "passthru_user1", "pass": "shortpin"}
        response = self.app.get(url(controller='validate', action='check'),
                                                            params=parameters)
        self.assertTrue('"value": true' in response, response)
        self.removeTokenBySerial(serial)


        ## with otppin==1 the pin should be the same
        self.setPinPolicy(realm='myDefRealm')
        serial = self.createToken(pin="otppin", typ='spass',
                                  user='passthru_user1')

        parameters = {"user": "passthru_user1", "pass": "geheim1"}
        response = self.app.get(url(controller='validate', action='check'),
                                params=parameters)
        self.assertTrue('"value": true' in response, response)
        self.removeTokenBySerial(serial)
        self.delPolicy()

        ## with otppin==2 the pin is not required at all
        self.setPinPolicy(realm='myDefRealm', action='otppin=2, ')
        serial = self.createToken(pin="otppin", typ='spass',
                                  user='passthru_user1')

        parameters = {"user": "passthru_user1", "pass": ""}
        response = self.app.get(url(controller='validate', action='check'),
                                params=parameters)
        self.assertTrue('"value": true' in response, response)
        self.removeTokenBySerial(serial)
        self.delPolicy()

        return

    def test_09_remote_regression(self):
        '''
        Challenge Response Test: regression remoteToken can split passw localy or remote
        '''

        serials = self.setup_remote_token()

        parameters = {"user": "remoteuser", "pass": "rpin123456"}
        response = self.app.get(url(controller='validate', action='check'),
                                params=parameters)
        
        self.assertTrue('"value": true' in response, response)

        parameters = {"user": "localuser", "pass": "lpin123456"}
        response = self.app.get(url(controller='validate', action='check'),
                                params=parameters)

        self.assertTrue('"value": true' in response, response)

        for serial in serials:
            self.removeTokenBySerial(serial)
        return


    def test_11_hmac_challenge_otppin1(self):
        '''
        Challenge Response Test: test hmac token with otppin=1 and challenge response
        '''

        counter = 0
        otpkey = "AD8EABE235FC57C815B26CEF3709075580B44738"

        ## with otppin==1 the pin should be the same
        self.setPinPolicy(realm='myDefRealm', action='otppin=1, ')

        serial = self.createToken(pin="otppin", user='passthru_user1')


        otp = self.calcOTP(otpkey, counter=counter)
        parameters = {"user": "passthru_user1", "pass": "geheim1" + otp}
        response = self.app.get(url(controller='validate', action='check'),
                                params=parameters)
        self.assertTrue('"value": true' in response, response)
        self.removeTokenBySerial(serial)

        self.delPolicy()

        ## with otppin==2 the pin is not required at all
        self.setPinPolicy(realm='myDefRealm', action='otppin=2, ')

        otp = self.calcOTP(otpkey, counter=counter)
        serial = self.createToken(pin="otppin", user='passthru_user1')
        parameters = {"user": "passthru_user1", "pass": otp}
        response = self.app.get(url(controller='validate', action='check'),
                                params=parameters)
        self.assertTrue('"value": true' in response, response)
        self.removeTokenBySerial(serial)

        self.delPolicy()

        return

    def test_01_hmac_challenge_std(self):
        '''
        Challenge Response Test: test if HMAC tokens still work - this is a potential challenge token
        '''
        counter = 0
        otpkey = "AD8EABE235FC57C815B26CEF3709075580B44738"
        ## normal test
        serial = self.createToken(pin="shortpin", otpkey=otpkey,
                                                    user='passthru_user1')

        ## submit a pin only request - to trigger a challenge
        parameters = {"user": "passthru_user1", "pass": "shortpin"}
        response = self.app.get(url(controller='validate', action='check'),
                                                            params=parameters)
        self.assertTrue('"value": false' in response, response)
        self.assertTrue('"transactionid":' not in response, response)

        self.setPinPolicy(name="ch_resp", realm='myDefRealm',
                                            action='challenge_response=hmac, ')

        ## submit a pin only request - to trigger a challenge
        parameters = {"user": "passthru_user1", "pass": "shortpin"}
        response = self.app.get(url(controller='validate', action='check'),
                                                            params=parameters)
        self.assertTrue('"value": false' in response, response)
        self.assertTrue('"transactionid":' in response, response)


        ## in the response we expect an transaction reference (=state)
        ## and a reply message message
        body = json.loads(response.body)
        try:
            state = body.get('detail').get('transactionid')
        except Exception as exx:
            raise Exception("missing challenge in body: %r, %r" % (body, exx))



        ## submit a otp only challenge response
        otp = self.calcOTP(otpkey, counter=counter)
        parameters = {"user": "passthru_user1", "pass": otp}
        parameters['transactionid'] = state
        response = self.app.get(url(controller='validate', action='check'),
                                                            params=parameters)
        self.assertTrue('"value": true' in response, response)


        ## submit a pin only request - to trigger a challenge
        parameters = {"user": "passthru_user1", "pass": "shortpin"}
        response = self.app.get(url(controller='validate', action='check'),
                                                            params=parameters)
        self.assertTrue('"value": false' in response, response)
        self.assertTrue('"transactionid":' in response, response)

        ## in the response we expect an transaction reference (=state)
        ## and a reply message message
        body = json.loads(response.body)
        state = body.get('detail').get('transactionid')

        ## submit a pin + otp challenge response
        counter = counter + 1
        otp = self.calcOTP(otpkey, counter=counter)
        parameters = {"user": "passthru_user1", "pass": otp}
        parameters['transactionid'] = state
        response = self.app.get(url(controller='validate', action='check'),
                                                            params=parameters)
        self.assertTrue('"value": true' in response, response)

        ## now create two open challenges

        ## submit a pin only request - to trigger a challenge
        parameters = {"user": "passthru_user1", "pass": "shortpin"}
        response = self.app.get(url(controller='validate', action='check'),
                                                            params=parameters)
        self.assertTrue('"value": false' in response, response)
        self.assertTrue('"transactionid":' in response, response)

        ## in the response we expect an transaction reference (=state)
        ## and a reply message message
        body = json.loads(response.body)
        state = body.get('detail').get('transactionid')

        ## submit a pin only request - to trigger a challenge
        parameters = {"user": "passthru_user1", "pass": "shortpin"}
        response = self.app.get(url(controller='validate', action='check'),
                                                            params=parameters)
        self.assertTrue('"value": false' in response, response)
        self.assertTrue('"transactionid":' in response, response)

        ## in the response we expect an transaction reference (=state)
        ## and a reply message message
        body = json.loads(response.body)
        state = body.get('detail').get('transactionid')


        ## submit a pin + otp challenge response
        counter = counter + 1
        otp = self.calcOTP(otpkey, counter=counter)
        parameters = {"user": "passthru_user1", "pass": otp}
        parameters['transactionid'] = state
        response = self.app.get(url(controller='validate', action='check'),
                                                            params=parameters)
        self.assertTrue('"value": true' in response, response)



        self.removeTokenBySerial(serial)

        self.delPolicy(name="ch_resp")
        return

    def test_10_multiple_tokens(self):
        """
        Challenge Response Test: authetication of multiple tokens using the transactionid

        Remark:
            with the hmac token, the transaction is not bound to a counter!
        """

        counter = 0
        otpkey = "AD8EABE235FC57C815B26CEF3709075580B44738"

        self.setPinPolicy(name="ch_resp", realm='myDefRealm',
                                       action='challenge_response=hmac topt,')

        _serial = self.createToken(serial="H1", pin="h1", otpkey=otpkey,
                                  user='passthru_user1', typ='hmac')

        _serial = self.createToken(serial="H2", pin="h2", otpkey=otpkey,
                                  user='passthru_user1', typ='hmac')

        ## submit a pin only request - to trigger a challenge
        for _i in range(1, 3):

            parameters = {"user": "passthru_user1", "pass": "h1"}
            response = self.app.get(url(controller='validate',
                                        action='check'), params=parameters)

            self.assertTrue('"value": false' in response, response)
            self.assertTrue('"transactionid":' in response, response)


            ## submit a pin only request - to trigger a challenge
            parameters = {"user": "passthru_user1", "pass": "h2"}
            response2 = self.app.get(url(controller='validate',
                                         action='check'), params=parameters)

            self.assertTrue('"value": false' in response2, response2)
            self.assertTrue('"transactionid":' in response2, response2)

            counter = counter + 1

        ## in the response we expect an transaction reference (=state)
        body = json.loads(response.body)
        state1 = body.get('detail').get('transactionid')

        body = json.loads(response2.body)
        state2 = body.get('detail').get('transactionid')

        #have a look, if all challenges are removed
        parameters = {"user": "passthru_user1"}
        response = self.app.get(url(controller='admin', action='checkstatus'),
                                                            params=parameters)

        self.assertTrue(state2 in response, response)

        #now check if the challenge could be identified by the last transaction
        otp = self.calcOTP(otpkey, counter=counter - 1)
        parameters = {"user": "passthru_user1", "pass": otp, 'state':state2}
        parameters['transactionid'] = state2
        response = self.app.get(url(controller='validate', action='check'),
                                                            params=parameters)
        self.assertTrue('"value": true' in response, response)

        #have a look, if all challenges are removed
        parameters = {"user": "passthru_user1"}
        response = self.app.get(url(controller='admin', action='checkstatus'),
                                                            params=parameters)

        self.assertTrue(state2 not in response, response)

        #reusage of the challenge should not work
        otp = self.calcOTP(otpkey, counter=counter)
        parameters = {"user": "passthru_user1", "pass": otp, 'state':state2}
        parameters['transactionid'] = state2
        response = self.app.get(url(controller='validate', action='check'),
                                                            params=parameters)

        self.assertTrue('"value": false' in response, response)

        #but the challenge for the other token should still be valid
        otp = self.calcOTP(otpkey, counter=counter)
        parameters = {"user": "passthru_user1", "pass": otp, 'state':state1}
        parameters['transactionid'] = state1
        response = self.app.get(url(controller='validate', action='check'),
                                                            params=parameters)

        self.assertTrue('"value": true' in response, response)

        #have a look, if all challenges are removed
        parameters = {"user": "passthru_user1"}
        response = self.app.get(url(controller='admin', action='checkstatus'),
                                                            params=parameters)

        ## assure that all challenges are removed
        self.assertTrue(state2 not in response, response)
        self.assertTrue(state1 not in response, response)


        self.removeTokenBySerial("H1")
        self.removeTokenBySerial("H2")

        self.delPolicy(name="ch_resp")
        return


    def test_12_sms_otppin(self):
        '''
        Challenge Response Test: SMS token challenge with otppin=1
        '''
        parameters = {
                'SMSProvider' : 'privacyidea.smsprovider.HttpSMSProvider.HttpSMSProvider',
                'selftest_admin' : 'superadmin'
                   }
        _response = self.app.get(url(controller='system', action='setConfig'),
                                                             params=parameters)

        sms_conf = { "URL" : "http://localhost:5001/testing/http2sms",
                     "PARAMETER" : { "account" : "clickatel",
                                    "username" : "legit" },
                    "SMS_TEXT_KEY":"text",
                    "SMS_PHONENUMBER_KEY":"destination",
                    "HTTP_Method" : "GET",
                    "RETURN_SUCCESS" : "ID"
                    }

        parameters = { 'SMSProviderConfig' : json.dumps(sms_conf),
                       'selftest_admin' : 'superadmin'
                      }

        response = self.app.get(url(controller='system', action='setConfig'),
                                params=parameters)
        log.error(response)
        self.assertTrue('"status": true' in response, response)

        # Patch (replace) smsprovider.HttpSMSProvider.HttpSMSProvider class
        # class to prevent sms from being sent out
        self.patch_sms = patch('privacyidea.smsprovider.HttpSMSProvider.HttpSMSProvider',
                             spec=privacyidea.smsprovider.HttpSMSProvider.HttpSMSProvider)
        mock_sms_class = self.patch_sms.start()
        mock_sms_instance = mock_sms_class.return_value
        mock_sms_instance.submitMessage.return_value = True
        mock_obj = mock_sms_instance.submitMessage

        typ = "sms"
        counter = 0
        otpkey = "AD8EABE235FC57C815B26CEF3709075580B44738"

        ## normal test
        serial = self.createToken(pin="shortpin", typ='sms', phone='12345',
                                        otpkey=otpkey, user='passthru_user1')

        ## trigger challenge
        parameters = {"user": "passthru_user1", "pass": "shortpin"}
        response = self.app.get(url(controller='validate', action='check'),
                                                            params=parameters)

        self.assertTrue('"value": false' in response, response)

        otp = get_otp(counter, otpkey, mock_obj, sms_otp_func, typ)
        parameters = {"user": "passthru_user1", "pass": "shortpin" + otp}
        response = self.app.get(url(controller='validate', action='check'),
                                params=parameters)

        self.assertTrue('"value": true' in response, response)

        ## now test same with otppin policy
        self.setPinPolicy(realm='myDefRealm', action='otppin=1, ')


        ## submit a pin only request - to trigger a challenge
        parameters = {"user": "passthru_user1", "pass": "geheim1"}
        response = self.app.get(url(controller='validate', action='check'),
                                                            params=parameters)

        self.assertTrue('"value": false' in response, response)

        ## validate sms
        otp = get_otp(counter, otpkey, mock_obj, sms_otp_func, typ)
        parameters = {"user": "passthru_user1", "pass": "geheim1" + otp}
        response = self.app.get(url(controller='validate', action='check'),
                                params=parameters)
        self.assertTrue('"value": true' in response, response)

        self.removeTokenBySerial(serial)
        self.delPolicy()

        return

    def test_16_remote_token_remote_challenge(self):
        '''
        Challenge Response Test: remoteToken with remote pin check
        '''
        counter = 0
        otpkey = "AD8EABE235FC57C815B26CEF3709075580B44738"
        user = "remoteuser"
        remoteurl = "http://127.0.0.1:5001"

        ## setup the remote token pairs
        serials = self.setup_remote_token(typ="hmac", otpkey=otpkey,
                                          remoteurl=remoteurl)

        ## now switch policy on for challenge_response for hmac token
        response = self.setPinPolicy(name="ch_resp", realm='*',
                                action='challenge_response=hmac remote')
        self.assertTrue('"status": true,' in response, response)

        response = self.setPinPolicy(name="ch_resp", realm='*',
                                action='challenge_response=hmac remote',
                                remoteurl=remoteurl)
        self.assertTrue('"status": true,' in response, response)

        # 1. part - pin belongs to remote token
        ## check is simple auth works
        otp = calcOTP(key=otpkey, counter=counter, typ="hmac")
        parameters = {"user": user, "pass": "rpin" + otp}
        response = self.app.get(url(controller='validate', action='check'),
                                params=parameters)

        self.assertTrue('"value": true' in response, response)

        # 1.1 now trigger a challenge
        otp = calcOTP(key=otpkey, counter=counter + 1, typ="hmac")
        parameters = {"user": user, "pass": "rpin"}
        response = self.app.get(url(controller='validate', action='check'),
                                params=parameters)

        self.assertTrue('"value": false' in response, response)

        body = json.loads(response.body)
        state = body.get('detail', {}).get('transactionid', '')
        self.assertTrue(state != '', response)

        # 1.2 check the challenge
        otp = calcOTP(key=otpkey, counter=counter + 1, typ="hmac")
        parameters = {"user": user, "pass": otp, "state" : state}
        response = self.app.get(url(controller='validate', action='check'),
                                params=parameters)
        ## hey, if this ok, we are done for the remote pin check
        self.assertTrue('"value": true' in response, response)

        for serial in serials:
            self.removeTokenBySerial(serial)

        self.delPolicy(name="ch_resp")
        self.delPolicy(name="ch_resp", remoteurl=remoteurl)

        return

    def test_17_remote_token_local_challenge(self):
        '''
        Challenge Response Test: remoteToken with with local pin check
        '''
        counter = 0
        otpkey = "AD8EABE235FC57C815B26CEF3709075580B44738"
        user = "localuser"
        remoteurl = "http://127.0.0.1:5001"

        ## setup the remote token pairs
        serials = self.setup_remote_token(typ="hmac",
                                          otpkey=otpkey,
                                          remoteurl=remoteurl)

        ## now switch policy on for challenge_response for hmac token
        response = self.setPinPolicy(name="ch_resp",
                                     realm='*',
                                     action='challenge_response=hmac remote')
        self.assertTrue('"status": true,' in response, response)

        response = self.setPinPolicy(name="ch_resp",
                                     realm='*',
                                     action='challenge_response=hmac remote',
                                     remoteurl=remoteurl)

        # now we have to test the local pin
        # when using the local pin, we will keep the challenge in the
        # src token

        # 1. part - pin belongs to local token - remote has no pin
        ## check is simple auth works
        otp = calcOTP(key=otpkey, counter=counter, typ="hmac")
        parameters = {"user": user, "pass": "lpin" + otp}
        response = self.app.get(url(controller='validate', action='check'),
                                params=parameters)

        self.assertTrue('"value": true' in response, response)

        # 2.1 now trigger a challenge
        counter = counter + 1
        otp = calcOTP(key=otpkey, counter=counter, typ="hmac")
        parameters = {"user": user, "pass": "lpin"}
        response = self.app.get(url(controller='validate', action='check'),
                                params=parameters)

        self.assertTrue('"value": false' in response, response)

        body = json.loads(response.body)
        state = body.get('detail', {}).get('transactionid', '')
        self.assertTrue(state != '', response)

        # 2.2 check the challenge
        counter = counter + 1
        otp = calcOTP(key=otpkey, counter=counter, typ="hmac")
        parameters = {"user": user, "pass": otp, "state" : state}
        response = self.app.get(url(controller='validate', action='check'),
                                params=parameters)
        ## hey, if this ok, we are done for the remote pin check
        self.assertTrue('"value": true' in response, response)

        for serial in serials:
            self.removeTokenBySerial(serial)

        self.delPolicy(name="ch_resp")
        self.delPolicy(name="ch_resp", remoteurl=remoteurl)

        return

    def test_22_radiustoken_remote_pin(self):
        """
        Challenge Response Test: radius token with remote PIN
        """
        # FIXME: we are missing the dummy radius server! :-(
        return
        serials = self.setup_radius_token()

        self._start_radius_server()
        user = "remoteuser"
        otp = "test123456"

        ## now switch policy on for challenge_response for hmac token
        response = self.setPinPolicy(name="ch_resp", realm='*',
                                action='challenge_response=radius')
        self.assertTrue('"status": true,' in response, response)

        # 1.1 now trigger a challenge
        parameters = {"user": user, "pass": "test"}
        response = self.app.get(url(controller='validate', action='check'),
                                params=parameters)

        self.assertTrue('"value": false' in response, response)

        body = json.loads(response.body)
        state = body.get('detail', {}).get('transactionid', '')
        self.assertTrue(state != '', response)

        # 1.2 check the challenge
        parameters = {"user": user, "pass": otp, "state" : state}
        response = self.app.get(url(controller='validate', action='check'),
                                params=parameters)
        ## hey, if this ok, we are done for the remote pin check
        self.assertTrue('"value": true' in response, response)

        self._stop_radius_server()

        for serial in serials:
            self.removeTokenBySerial(serial)

        return

    def test_23_radiustoken_local_pin(self):
        """
        Challenge Response Test: radius token with local PIN
        """
        # FIXME: we are missing the dummy radius server! :-(
        return
        serials = self.setup_radius_token()
        self._start_radius_server()

        user = "localuser"
        otp = "654321"

        ## now switch policy on for challenge_response for hmac token
        response = self.setPinPolicy(name="ch_resp", realm='*',
                                action='challenge_response=radius')
        self.assertTrue('"status": true,' in response, response)

        # 1.1 now trigger a challenge
        parameters = {"user": user, "pass": "local"}
        response = self.app.get(url(controller='validate', action='check'),
                                params=parameters)

        self.assertTrue('"value": false' in response, response)

        body = json.loads(response.body)
        state = body.get('detail', {}).get('transactionid', '')
        self.assertTrue(state != '', response)

        # 1.2 check the challenge
        parameters = {"user": user, "pass": otp, "state" : state}
        response = self.app.get(url(controller='validate', action='check'),
                                params=parameters)
        ## hey, if this ok, we are done for the remote pin check
        self.assertTrue('"value": true' in response, response)

        self._stop_radius_server()

        for serial in serials:
            self.removeTokenBySerial(serial)

        return

    def do_auth(self, pin='',
                    counter=0,
                    otpkey="AD8EABE235FC57C815B26CEF3709075580B44738",
                    typ='hmac',
                    mock_obj=None,
                    otp_func=None,
                    user='passthru_user1',
                    ):
        """
        run a set of different authentication schemes:
        * std auth with pin+otp
        * challenge + response w. pin+otp
        * challenge + response w. transid+otp

        :param pin: the pin, depending on otppin policy: pin/pass/empty
        :param counter: the counter increment to provide the correct otp
        :param otpkey: the key to calculate the next otp

        :return: the last otpcount to continue authentication
        """


        if typ not in  ['email', 'sms']:
            #1 std auth with user with pin+otp
            counter = counter + 1
            otp = self.calcOTP(otpkey, counter=counter, typ=typ)
            if typ == "totp":
                time.sleep(TOTP_TIMESTEP)
            parameters = {"user": user, "pass": pin + otp}
            response = self.app.get(url(controller='validate', action='check'),
                                                            params=parameters)
            self.assertTrue('"value": true' in response, response)

        #2. challenge response with pin+otp
        #2.1. challenge
        parameters = {"user": user, "pass": pin, }
        response = self.app.get(url(controller='validate', action='check'),
                                                            params=parameters)
        self.assertTrue('"value": false' in response, response)

        #2.2 response
        counter = counter + 1
        otp = get_otp(counter, otpkey, mock_obj, otp_func, typ=typ)
        if typ == "totp":
            time.sleep(TOTP_TIMESTEP)        
        parameters = {"user": user, "pass": pin + otp}
        response = self.app.get(url(controller='validate', action='check'),
                                                            params=parameters)

        self.assertTrue('"value": true' in response, response)

        #3. challenge response with otp+state
        #3.1 trigger challenge
        parameters = {"user": user, "pass": pin}
        response = self.app.get(url(controller='validate', action='check'),
                                                            params=parameters)

        self.assertTrue('"value": false' in response, response)

        body = json.loads(response.body)
        state = body.get('detail').get('transactionid')

        #3.2 response
        counter = counter + 1
        otp = get_otp(counter, otpkey, mock_obj, otp_func, typ=typ)
        if typ == "totp": 
            time.sleep(TOTP_TIMESTEP) 
        parameters = {"user": user, "pass": otp, "state" : state}
        response = self.app.get(url(controller='validate', action='check'),
                                                            params=parameters)

        self.assertTrue('"value": true' in response, response)

        #4 std auth with user with pin+otp though outstanding challenge
        #4.1 trigger challenge
        parameters = {"user": user, "pass": pin}
        response = self.app.get(url(controller='validate', action='check'),
                                                            params=parameters)

        self.assertTrue('"value": false' in response, response)

        #4.2 do std auth
        counter = counter + 1
        otp = get_otp(counter, otpkey, mock_obj, otp_func, typ=typ)
        if typ == "totp": 
            time.sleep(TOTP_TIMESTEP)
        parameters = {"user": user, "pass": pin + otp}
        response = self.app.get(url(controller='validate', action='check'),
                                                            params=parameters)

        self.assertTrue('"value": true' in response, response)

        return counter

    def test_50_hmac_auth(self):
        '''
        Challenge Response Test: hmac token challenge with otppin=1 + otppin=2
        '''

        counter = 0
        otpkey = "AD8EABE235FC57C815B26CEF3709075580B44738"

        serial = self.createToken(pin="shortpin", typ='hmac',
                                        otpkey=otpkey, user='passthru_user1')

        ## now switch policy on for challenge_response
        response = self.setPinPolicy(name="ch_resp", realm='myDefRealm',
                                action='challenge_response=hmac,')
        self.assertTrue('"status": true,' in response, response)

        counter = self.do_auth("shortpin", counter)

        ## with otppin==1 the pin should be the same as the password
        response = self.setPinPolicy(realm='myDefRealm', action='otppin=1, ')
        self.assertTrue('"status": true,' in response, response)

        counter = self.do_auth("geheim1", counter + 1)

        ## with otppin==2 the pin should be the same as the password
        response = self.setPinPolicy(realm='myDefRealm', action='otppin=2, ')
        self.assertTrue('"status": true,' in response, response)

        counter = self.do_auth("", counter + 1)

        self.removeTokenBySerial(serial)
        self.delPolicy(name="ch_resp")
        self.delPolicy(name="otpPin")

        return


    def test_51_sms_auth(self):
        '''
        Challenge Response Test: sms token challenge with otppin=1 + otppin=2
        '''
        if not SMS:
            return
        typ = 'sms'
        parameters = {
                'SMSProvider' : 'privacyidea.smsprovider.HttpSMSProvider.HttpSMSProvider',
                'selftest_admin' : 'superadmin'
                   }
        _response = self.app.get(url(controller='system', action='setConfig'),
                                                             params=parameters)

        sms_conf = { "URL" : "http://localhost:5001/testing/http2sms",
                     "PARAMETER" : { "account" : "clickatel",
                                    "username" : "legit" },
                    "SMS_TEXT_KEY":"text",
                    "SMS_PHONENUMBER_KEY":"destination",
                    "HTTP_Method" : "GET",
                    "RETURN_SUCCESS" : "ID"
                    }

        parameters = { 'SMSProviderConfig' : json.dumps(sms_conf),
                       'selftest_admin' : 'superadmin'
                      }

        response = self.app.get(url(controller='system', action='setConfig'),
                                params=parameters)
        log.error(response)
        self.assertTrue('"status": true' in response, response)

        counter = 0
        otpkey = "AD8EABE235FC57C815B26CEF3709075580B44738"

        serial = self.createToken(pin="shortpin", typ='sms', phone="123456",
                                        otpkey=otpkey, user='passthru_user1')

        # Patch (replace) smsprovider.HttpSMSProvider.HttpSMSProvider class
        # class to prevent sms from being sent out
        self.patch_sms = patch('privacyidea.smsprovider.HttpSMSProvider.HttpSMSProvider',
                             spec=privacyidea.smsprovider.HttpSMSProvider.HttpSMSProvider)
        mock_sms_class = self.patch_sms.start()
        mock_sms_instance = mock_sms_class.return_value
        mock_sms_instance.submitMessage.return_value = True
        mock_obj = mock_sms_instance.submitMessage

        ## sms token should do challenge response even without policy
        counter = self.do_auth("shortpin", counter, typ=typ,
                                            mock_obj=mock_obj,
                                            otp_func=sms_otp_func
                                            )

        ## now switch policy on for challenge_response - should not harm
        response = self.setPinPolicy(name="ch_resp", realm='myDefRealm',
                                action='challenge_response=hmac topt sms,')
        self.assertTrue('"status": true,' in response, response)

        counter = self.do_auth("shortpin", counter, typ=typ,
                                            mock_obj=mock_obj,
                                            otp_func=sms_otp_func
                                            )

        ## with otppin==1 the pin should be the same as the password
        response = self.setPinPolicy(realm='myDefRealm', action='otppin=1, ')
        self.assertTrue('"status": true,' in response, response)

        counter = self.do_auth("geheim1", counter, typ=typ,
                                            mock_obj=mock_obj,
                                            otp_func=sms_otp_func
                                            )

        ## with otppin==2 the pin should be the same as the password
        response = self.setPinPolicy(realm='myDefRealm', action='otppin=2, ')
        self.assertTrue('"status": true,' in response, response)

        counter = self.do_auth("", counter, typ=typ,
                                            mock_obj=mock_obj,
                                            otp_func=sms_otp_func
                                            )
        self.removeTokenBySerial(serial)
        self.delPolicy(name="ch_resp")
        self.delPolicy(name="otpPin")

        return


    def test_52_email_auth(self):
        """
        Challenge Response Test: email token challenge with otppin=1 + otppin=2
        """
        typ = 'email'

        parameters = {
            'EmailProvider': 'privacyidea.lib.emailprovider.SMTPEmailProvider',
            'EmailProviderConfig': '{ "SMTP_SERVER": "mail.example.com",\
                               "SMTP_USER": "secret_user",\
                               "SMTP_PASSWORD": "secret_pasword" }',
            'EmailChallengeValidityTime': 300,
            'EmailBlockingTimeout': 0
        }
        response = self.app.get(url(controller='system', action='setConfig'),
                                params=parameters)
        self.assertTrue('"status": true' in response, response)

        # Patch (replace) smtplib.SMTP class to prevent e-mails from being sent out
        self.patch_smtp = patch('smtplib.SMTP', spec=smtplib.SMTP)
        mock_smtp_class = self.patch_smtp.start()
        mock_smtp_instance = mock_smtp_class.return_value
        mock_smtp_instance.sendmail.return_value = []
        mock_obj = mock_smtp_instance.sendmail

        # Enroll token
        pin = "shortpin"
        _description = "E-mail token enrolled in functional tests",
        otpkey = "AD8EABE235FC57C815B26CEF3709075580B44738"

        _serial = self.createToken(pin=pin,
                                typ=typ,
                                email_address='paul@example.com',
                                description="email token",
                                otpkey=otpkey,
                                user='passthru_user1')

        counter = 0

        ## email token should do challenge response even without policy
        counter = self.do_auth("shortpin", counter, typ=typ,
                                            mock_obj=mock_obj,
                                            otp_func=email_otp_func
                                            )

        ## now switch policy on for challenge_response - should not harm
        response = self.setPinPolicy(name="ch_resp", realm='myDefRealm',
                                action='challenge_response=hmac email,')
        self.assertTrue('"status": true,' in response, response)

        counter = self.do_auth("shortpin", counter, typ=typ,
                                            mock_obj=mock_obj,
                                            otp_func=email_otp_func
                                            )

        ## with otppin==1 the pin should be the same as the password
        response = self.setPinPolicy(realm='myDefRealm',
                                                    action='otppin=1, ')
        self.assertTrue('"status": true,' in response, response)

        counter = self.do_auth("geheim1", counter, typ=typ,
                                            mock_obj=mock_obj,
                                            otp_func=email_otp_func
                                            )

        ## with otppin==2 the pin should be the same as the password
        response = self.setPinPolicy(realm='myDefRealm',
                                                    action='otppin=2, ')
        self.assertTrue('"status": true,' in response, response)

        counter = self.do_auth("", counter, typ=typ,
                                            mock_obj=mock_obj,
                                            otp_func=email_otp_func
                                            )

        return

    def test_54_totp_auth(self):
        '''
        Challenge Response Test: totp token challenge with otppin=1 + otppin=2
        '''
    
        typ = 'totp'
        counter = 0
        otpkey = "AD8EABE235FC57C815B26CEF3709075580B44738"

        serial = self.createToken(pin="shortpin", typ=typ,
                                        otpkey=otpkey, user='passthru_user1')

        ## now switch policy on for challenge_response
        response = self.setPinPolicy(name="ch_resp", realm='myDefRealm',
                                action='challenge_response=hmac totp,')
        self.assertTrue('"status": true,' in response, response)

        counter = self.do_auth("shortpin", counter, typ=typ)

        # reduce the test time
        return

        ## with otppin==1 the pin should be the same as the password
        response = self.setPinPolicy(realm='myDefRealm', action='otppin=1, ')
        self.assertTrue('"status": true,' in response, response)


        counter = self.do_auth("geheim1", counter + 1, typ=typ)

        ## with otppin==2 the pin should be the same as the password
        response = self.setPinPolicy(realm='myDefRealm', action='otppin=2, ')
        self.assertTrue('"status": true,' in response, response)

        counter = self.do_auth("", counter + 1, typ=typ)

        self.removeTokenBySerial(serial)
        self.delPolicy(name="ch_resp")
        self.delPolicy(name="otpPin")

        return

    def test_60_hmac_two_tokens(self):
        '''
        Challenge Response Test: two hmac token in different realms

        test with
        * passthru_user1 - using default realm - using  token2
        * passthru_user1@myMixRealm .- using token1
        * passthru_user1@myDefRealm - using token2

        in combination with optpin=1 and otppin=2
        '''
        typ = "hmac"

        ## now switch policy on for challenge_response
        response = self.setPinPolicy(name="ch_resp1", realm='myDefRealm',
                                action='challenge_response=hmac,')
        self.assertTrue('"status": true,' in response, response)

        response = self.setPinPolicy(name="ch_resp2", realm='myMixRealm',
                                action='challenge_response=hmac,')
        self.assertTrue('"status": true,' in response, response)

        counter1 = 0
        counter2 = 0

        otpkey1 = "AD8EABE235FC57C815B26CEF3709075580B44738"
        otpkey2 = "38AD8EABE235FC57C815B26CEF3709075580B447"

        _serial1 = self.createToken(serial='myMixRealm', pin="shortpin",
                        typ=typ, otpkey=otpkey1,
                        user='passthru_user1@myMixRealm')

        _serial2 = self.createToken(serial='myDefRealm', pin="shortpin",
                        typ=typ, otpkey=otpkey2,
                        user='passthru_user1@myDefRealm')

        ## submit a pin only request - to trigger a challenge
        parameters = {"user": "passthru_user1", "pass": "shortpin"}
        response = self.app.get(url(controller='validate', action='check'),
                                                            params=parameters)
        self.assertTrue('"value": false' in response, response)
        self.assertTrue('"transactionid":' in response, response)

        ## in the response we expect a transaction reference (=state)
        ## and a reply message
        body = json.loads(response.body)
        state = body.get('detail', {}).get('transactionid', '')
        self.assertTrue(len(state) > 0, body)

        counter2 = counter2 + 1
        otp = get_otp(counter2, otpkey2, None, None, typ)
        ## submit a pin only request - to trigger a challenge
        parameters = {"user": "passthru_user1", "pass": otp, "state":state}
        response = self.app.get(url(controller='validate', action='check'),
                                                            params=parameters)

        self.assertTrue('"value": true' in response, response)

        ## submit a pin only request - to trigger a challenge
        parameters = {"user": "passthru_user1@myMixRealm", "pass": "shortpin"}
        response = self.app.get(url(controller='validate', action='check'),
                                                            params=parameters)
        self.assertTrue('"value": false' in response, response)
        self.assertTrue('"transactionid":' in response, response)

        ## in the response we expect a transaction reference (=state)
        ## and a reply message
        body = json.loads(response.body)
        state = body.get('detail', {}).get('transactionid', '')
        self.assertTrue(len(state) > 0, body)

        counter1 = counter1 + 1
        otp = get_otp(counter1, otpkey1, None, None, typ)
        ## submit a pin only request - to trigger a challenge
        parameters = {"user": "passthru_user1@myMixRealm",
                                                "pass": otp, "state":state}
        response = self.app.get(url(controller='validate', action='check'),
                                                            params=parameters)

        self.assertTrue('"value": true' in response, response)

        ## submit a pin only request - to trigger a challenge
        parameters = {"user": "passthru_user1@myDefRealm", "pass": "shortpin"}
        response = self.app.get(url(controller='validate', action='check'),
                                                            params=parameters)
        self.assertTrue('"value": false' in response, response)
        self.assertTrue('"transactionid":' in response, response)

        ## in the response we expect a transaction reference (=state)
        ## and a reply message
        body = json.loads(response.body)
        state = body.get('detail', {}).get('transactionid', '')
        self.assertTrue(len(state) > 0, body)

        counter2 = counter2 + 1
        otp = get_otp(counter2, otpkey2, None, None, typ)
        ## submit a pin only request - to trigger a challenge
        parameters = {"user": "passthru_user1@myDefRealm",
                                                "pass": otp, "state":state}
        response = self.app.get(url(controller='validate', action='check'),
                                                            params=parameters)

        self.assertTrue('"value": true' in response, response)

        self.delPolicy()

        counter2 = self.do_auth("shortpin", counter2 + 1 , otpkey=otpkey2,
                                user="passthru_user1@myDefRealm")

        ## with otppin==1 the pin should be the same as the password
        response = self.setPinPolicy(realm='myDefRealm', action='otppin=1, ')
        self.assertTrue('"status": true,' in response, response)

        counter2 = self.do_auth("geheim1", counter2 + 1, otpkey=otpkey2,
                                user="passthru_user1@myDefRealm")

        ## with otppin==2 the pin should be the same as the password
        response = self.setPinPolicy(realm='myDefRealm', action='otppin=2, ')
        self.assertTrue('"status": true,' in response, response)

        counter2 = self.do_auth("", counter2 + 1, otpkey=otpkey2,
                               user="passthru_user1@myDefRealm")

        self.delPolicy()

        counter1 = self.do_auth("shortpin", counter1 + 1 , otpkey=otpkey1,
                                user="passthru_user1@myMixRealm")

        ## with otppin==1 the pin should be the same as the password
        response = self.setPinPolicy(realm='myMixRealm', action='otppin=1, ')
        self.assertTrue('"status": true,' in response, response)

        counter1 = self.do_auth("geheim1", counter1 + 1, otpkey=otpkey1,
                                user="passthru_user1@myMixRealm")

        ## with otppin==2 the pin should be the same as the password
        response = self.setPinPolicy(realm='myMixRealm', action='otppin=2, ')
        self.assertTrue('"status": true,' in response, response)

        counter1 = self.do_auth("", counter1 + 1, otpkey=otpkey1,
                               user="passthru_user1@myMixRealm")

        self.delPolicy()

        counter2 = self.do_auth("shortpin", counter2 + 1 , otpkey=otpkey2,
                                user="passthru_user1")

        ## with otppin==1 the pin should be the same as the password
        response = self.setPinPolicy(realm='myDefRealm', action='otppin=1, ')
        self.assertTrue('"status": true,' in response, response)

        counter2 = self.do_auth("geheim1", counter2 + 1, otpkey=otpkey2,
                                user="passthru_user1")

        ## with otppin==2 the pin should be the same as the password
        response = self.setPinPolicy(realm='myDefRealm', action='otppin=2, ')
        self.assertTrue('"status": true,' in response, response)

        counter2 = self.do_auth("", counter2 + 1, otpkey=otpkey2,
                               user="passthru_user1")


        return

    def test_61_hmac_active_inactive_tokens(self):
        '''
        Challenge Response Test: two hmac token (active/inactive) for one user

        tests fix for #12413
        '''
        typ = "hmac"
        otpkey1 = "AD8EABE235FC57C815B26CEF3709075580B44738"
        otpkey2 = "38AD8EABE235FC57C815B26CEF3709075580B447"

        ## now switch policy on for challenge_response
        response = self.setPinPolicy(name="ch_resp1", realm='myDefRealm',
                                action='challenge_response=hmac,')
        self.assertTrue('"status": true,' in response, response)

        _serial1 = self.createToken(serial='one',
                                   pin="shortpin",
                                   typ=typ,
                                   otpkey=otpkey1,
                                   description="one",
                                   user='passthru_user1@myDefRealm')

        _serial2 = self.createToken(serial='two',
                                   pin="shortpin2",
                                   typ=typ,
                                   otpkey=otpkey2,
                                   description="two",
                                   user='passthru_user1@myDefRealm')

        ## if the pin is a unique identifier a challenge should be triggered
        ## though there are two tokens belonging to one user and both are
        # active
        parameters = {"user": "passthru_user1", "pass": "shortpin"}
        response = self.app.get(url(controller='validate',
                                    action='check'),
                                params=parameters)

        self.assertTrue('"value": false' in response, response)
        self.assertTrue('"transactionid":' in response, response)

        serial2 = self.createToken(serial='two',
                                   pin="shortpin",
                                   typ=typ,
                                   otpkey=otpkey2,
                                   description="two",
                                   user='passthru_user1@myDefRealm')

        ## now the pin is not a unique identifier anymore and both tokens are
        ## active an error will return because multiple challenges could have
        ## been triggered
        parameters = {"user": "passthru_user1", "pass": "shortpin"}
        response = self.app.get(url(controller='validate', action='check'),
                                params=parameters)

        self.assertTrue('processing of multiple challenges is '
                        'not supported!' in response, response)

        parameters = {"serial": serial2}
        response = self.app.get(url(controller='admin', action='disable'),
                                params=parameters)

        self.assertTrue('"value": 1' in response, response)

        parameters = {"user": "passthru_user1", "pass": "shortpin"}
        response = self.app.get(url(controller='validate', action='check'),
                                params=parameters)

        self.assertTrue('"value": false' in response, response)
        self.assertTrue('"transactionid":' in response, response)

        return

    def test_62_email_active_inactive_tokens(self):
        '''
        Challenge Response Test: two email token (active/inactive) for one user

        tests fix for #12416
        '''
        typ = 'email'

        parameters = {
            'EmailProvider': 'privacyidea.lib.emailprovider.SMTPEmailProvider',
            'EmailProviderConfig': '{ "SMTP_SERVER": "mail.example.com",\
                               "SMTP_USER": "secret_user",\
                               "SMTP_PASSWORD": "secret_pasword" }',
            'EmailChallengeValidityTime': 300,
            'EmailBlockingTimeout': 0
        }
        response = self.app.get(url(controller='system', action='setConfig'),
                                params=parameters)
        self.assertTrue('"status": true' in response, response)

        # Patch (replace) smtplib.SMTP class to prevent
        # e-mails from being sent out
        self.patch_smtp = patch('smtplib.SMTP', spec=smtplib.SMTP)
        mock_smtp_class = self.patch_smtp.start()
        mock_smtp_instance = mock_smtp_class.return_value
        mock_smtp_instance.sendmail.return_value = []
        mock_obj = mock_smtp_instance.sendmail

        otpkey1 = "AD8EABE235FC57C815B26CEF3709075580B44738"
        otpkey2 = "38AD8EABE235FC57C815B26CEF3709075580B447"

        # create two email tokens
        _serial1 = self.createToken(serial='one',
                                    pin="shortpin",
                                    typ=typ,
                                    email_address='paul@example.com',
                                    otpkey=otpkey1,
                                    description="one",
                                    user='passthru_user1@myDefRealm')

        _serial2 = self.createToken(serial='two',
                                   pin="shortpin2",
                                   typ=typ,
                                   email_address='paul@example.com',
                                   otpkey=otpkey2,
                                   description="two",
                                   user='passthru_user1@myDefRealm')

        counter = 0
        ## email token should do challenge response even without policy
        counter = self.do_auth("shortpin", counter, typ=typ,
                               mock_obj=mock_obj, otp_func=email_otp_func,
                               user="passthru_user1")

        serial2 = self.createToken(serial='two',
                                   pin="shortpin",
                                   typ=typ,
                                   email_address='paul@example.com',
                                   otpkey=otpkey2,
                                   description="two",
                                   user='passthru_user1@myDefRealm')

        ## now the pin is not a unique identifier anymore and both tokens are
        ## active an error will return because multiple challenges could have
        ## been triggered
        parameters = {"user": "passthru_user1", "pass": "shortpin"}
        response = self.app.get(url(controller='validate', action='check'),
                                params=parameters)

        self.assertTrue('processing of multiple challenges is '
                        'not supported!' in response, response)

        parameters = {"serial": serial2}
        response = self.app.get(url(controller='admin', action='disable'),
                                params=parameters)

        self.assertTrue('"value": 1' in response, response)

        parameters = {"user": "passthru_user1", "pass": "shortpin"}
        response = self.app.get(url(controller='validate', action='check'),
                                params=parameters)

        self.assertTrue('"value": false' in response, response)
        self.assertTrue('"transactionid":' in response, response)

        ## email token should do challenge response even without policy
        counter = self.do_auth("shortpin", counter, typ=typ,
                               mock_obj=mock_obj, otp_func=email_otp_func,
                               user="passthru_user1")

        ## with otppin==1 the pin should be the same as the password
        response = self.setPinPolicy(realm='myDefRealm',
                                        action='otppin=1, ')
        self.assertTrue('"status": true,' in response, response)

        ## email token should do challenge response even with pin == passw
        counter = self.do_auth("geheim1", counter, typ=typ,
                               mock_obj=mock_obj, otp_func=email_otp_func,
                               user="passthru_user1")

        ## with otppin==2 the pin should be the same as the password
        response = self.setPinPolicy(realm='myDefRealm',
                                        action='otppin=2, ')
        self.assertTrue('"status": true,' in response, response)

        ## email token should do challenge response even with pin is none
        counter = self.do_auth("", counter, typ=typ,
                               mock_obj=mock_obj, otp_func=email_otp_func,
                               user="passthru_user1")

        return

    def test_63_sms_config_error(self):
        '''
        Challenge Response Test: sms token challenge and error in config
        '''
        typ = 'sms'
        parameters = {
                'SMSProvider' : 'smsprovider.SmtpSMSProvider.SmtpSMSProvider',
                'selftest_admin' : 'superadmin'
                   }
        _response = self.app.get(url(controller='system', action='setConfig'),
                                                             params=parameters)

        sms_conf = { "mailserver" : "somemailserverthatdoesntexist.com",
                      "mailsender" : "privacyidea-sms@lsexperts.de",
                      "mailto": "omar.kohl@lsexperts.de",
                      "mailuser": "myuser",
                      "mailpassword": "12345",
                      "subject" : "<phone>",
                      "body" : "This is your OTP-value: <otp>" }

        parameters = { 'SMSProviderConfig' : json.dumps(sms_conf),
                       'selftest_admin' : 'superadmin'
                      }

        response = self.app.get(url(controller='system', action='setConfig'),
                                params=parameters)
        log.error(response)
        self.assertTrue('"status": true' in response, response)

        otpkey = "AD8EABE235FC57C815B26CEF3709075580B44738"

        _serial = self.createToken(pin="shortpin", typ=typ, phone="123456",
                                        otpkey=otpkey, user='passthru_user1')

        # as the config contains an error, the result message should contain
        # a error message and not the otppin

        parameters = {"user": "passthru_user1", "pass": "shortpin"}
        response = self.app.get(url(controller='validate', action='check'),
                                params=parameters)

        self.assertTrue('the SMS could not be sent' in response, response)

        return

    def test_72_exception_in_challenge(self):
        '''
        Challenge Response Test: handle exception during create_challenge

        '''
        typ = 'email'

        parameters = {
            'EmailProvider': 'privacyidea.lib.emailprovider.SMTPEmailProvider',
            'EmailProviderConfig': '{ "SMTP_SERVER": "mailss.example.com",\
                               "SMTP_USER": "secret_user",\
                               "SMTP_PASSWORD": "secret_pasword" }',
            'EmailChallengeValidityTime': 300,
            'EmailBlockingTimeout': 0
        }
        response = self.app.get(url(controller='system', action='setConfig'),
                                params=parameters)
        self.assertTrue('"status": true' in response, response)

        otpkey1 = "AD8EABE235FC57C815B26CEF3709075580B44738"

        # create email tokens
        serial1 = self.createToken(serial='one',
                                    pin="shortpin",
                                    typ=typ,
                                    email_address='paul@example.com',
                                    otpkey=otpkey1,
                                    description="one",
                                    user='passthru_user1@myDefRealm')

        # does validate/check displays an error and the status false
        parameters = {"user": "passthru_user1", "pass": "shortpin"}
        response = self.app.get(url(controller='validate', action='check'),
                                params=parameters)
        self.assertTrue('"message": "validate/check failed:' in response,
                        response)
        self.assertTrue('"status": false' in response, response)


        # check if simplecheck displays as well an error
        parameters = {"user": "passthru_user1", "pass": "shortpin"}
        response = self.app.get(url(controller='validate',
                                    action='simplecheck'),
                                params=parameters)

        self.assertTrue(':-/' in response, response)

        # finally check, if there is no open challenge left
        parameters = {"serial": serial1}
        response = self.app.get(url(controller='admin',
                                    action='checkstatus'),
                                params=parameters)

        self.assertTrue('"values": {}' in response, response)

        return

##eof##########################################################################
